/*******************************************************************/
/*                                                                 */
/*   Program Name :  pcalcheck                                     */
/*                                                                 */
/*      Version:  1.00  2007-10-26                                 */
/*                                                                 */
/*      Copyright (c) 2007-     T.Kondo/NICT All Right Reserved    */
/*                                                                 */
/*******************************************************************/
/*
  K5/VSSP and K5/VSSP32 PCAL check
          T.Kondo   
          
            Ver 1.00 2007-10-25
            Ver 1.01 2007-10-26  leave graphic window at end of
                                 program in case of Windows
            Ver 1.10 2007-11-03  support GNUPLOT, however GNUPLOT is not
                                 good at real time plot.  Processing speed
                                 will reduce gradually at after several point plots.
                                 So recommend to use an option "-after" that enforces
                                 a plot after all data are processed 
                                 (this option is default for GNUPLOT default).
                                 use "-noafter" if you want to get a plot and update 
                                 with data processing.

                                 
                               


    Usage:
       pcalcheck k5file [option]  
        or
       pcalcheck ofile [-pgplot device | -ps]  --- for re-plot

         where  k5file  ---  K5/VSSP or K5/VSSP32 file name
                ofile   ---  pcalcheck output file

              [option]
                  -fall pcalf --- set PCAL frequency for all channels (kHz)
                                  (default 10kHz)
                  -f1 pcalf1 --- set PCAL frequency for CH1 (kHz)
                  -f2 pcalf2 --- set PCAL frequency for CH2 (kHz)
                  -f3 pcalf3 --- set PCAL frequency for CH3 (kHz)
                  -f4 pcalf4 --- set PCAL frequency for CH4 (kHz)
                  -integ integ --- set integration period (sec)
                                   (default is all scan period)
                  -pgplot device -- set PGPLOT device 
                                     /NULL for suppression
                                     /CPS for colored postscript
                                     (default is  display)
                  -ps  -- compulsory set graphic device to POSTSCRIPT
                                 (valid for re-plot mode)
                  -out ofile -- set output file name
                                 (default is "pcalcheck_out.txt")
                  -nocheck --- no header check
                      following parameter for nocheck mode
                  -vssp32  --- set VSSP32 mode
                  -sfreq sfreq --- set sampling frequency (MHz)
                  -adbit adbit --- set AD bits
                  -numch numch --- set # of channels
                  -after   --- plot graph after all data processing
                              (this is default in case of GNUPLOT)
                  -noafter  --- plot and update graph with data processing
                              (this is default in case of PGPLOT)

pcaldata.txt                                    
D:\IPVLBI\data\k5vssp32\VSSP32sample.dat -f3 990 -f4 990 -integ 2 -after
*/
#define PROGRAM_VERSION "Ver. 2007-11-04"

#define PCALCHECK_OUT_DFLT "pcalcheck_out.txt"    // pcal detection results out put file

//#define USE_GNUPLOT    // use GNUPLOT  (default is use of PGPLOT)

#include "g_head.h"    // header for graphics

#ifdef WIN32
  int tempcounter=0;  // counter for tempfile manage
#endif


#include "libk5.h"
//#include "cor_head.h"

#ifndef MAXPP
  #define MAXPP 2048
#endif

/* default number of channels */
#ifndef CHSIZE
  #define CHSIZE 4
#endif


//#define NOGRAPH    // in case of no graph output

/**** copy from libcor.c and libcor.h *****/

typedef struct pcaltype_v3    // pcal detection type definition (original is in libcor.h)
{
    double *pcalf;          // pcal frequency by channel size should be CHSIZE
    double *rpcal;    // real part pcal corr func
    double *ipcal;    // imaginary part pcal corr func
    double *sum1;     // for normalization
    double *sum2;     // for normalization
    int ns;    // # of samples processed at each PP
    int kpp;          // current PP index
    int mode;         // PCAL detection mode
                      // 0: no detection
                      // 1: PCAL detection on all pcal frequency is same as first ch
                      // 2: PCAL detection on pcal frequency is different by ch
    int tmode;        // PCAL detection integration mode
                      // 0: every 1 sec
                      // 1: same as PP period
    int numch;        // actual channel number  (1 or 4)
    int chsize;       // IP-BOARD apriori chnnel size (usually 4)
} pcaltype_tv3;


void setctabl0(int* cpat);
void setctabl(char* cT);
void inipcaldatv3(pcaltype_tv3 *pc);
void pcaldatnormv3(pcaltype_tv3 *pc);
void pcaldatnorm_1bit(pcaltype_tv3 *pc);
int checkheader(ip_vlbi_file_t *S);
int DatRead1bitSv(ip_vlbi_file_t *S, 
                 unsigned char* ibuf,
		         unsigned char *idata, 
                 int numb, 
                 pcaltype_tv3 *pc, int *cpat, char *pcalcT);
long int Dat1unitRead_w_Pcdetv3(ip_vlbi_file_t *S, unsigned char* ibuf,
					  int *idata, int numb, int *kpos, pcaltype_tv3 *pc);
void make_extended_filename(char *fname, char *fname2);


void setctabl0(int* cpat)
{
     int i,k,d,bsum;
	 // create 8bit pattern counts table 
	 for (i=0;i<=255; i++)
	 {
		d=i;
	    bsum=0;
		for (k=0;k<8;k++){
		  if( (d & 1)==1)
			 bsum++;
		     d=d >> 1;
		}
        cpat[i]=bsum;
	 }
}

void setctabl(char* cT)
{
     int i,j,k,m, a,b,c,d,bsum;
     int *ctemp;
	 ctemp = (int *)malloc(256*sizeof(int));
	 // create 8bit pattern counts table for temporary use
	 for (i=0;i<=255; i++)
	 {
		d=i;
	    bsum=0;
		for (k=0;k<8;k++){
		  if( (d & 1)==0)
			 bsum++;
		     d=d >> 1;
		}
        ctemp[i]=bsum;
	 }

	 for (i=0;i<=255; i++){
		 a=i;
		 for (j=0;j<=255; j++){
			 b=j;
			 m=i*256+j;
			 c= a ^ b;
			 /*
			// if (i==0 && j<10) cout << c <<"\n";
			 bsum=0;
			 for (k=0;k<=7;k++){
				 if( (c & 1)==0)
					 bsum++;
				 c=c >> 1;
			 }
            cT[m]=(char)bsum;   
			*/
			cT[m]=(char)ctemp[c];
		 }
	 }
	 free(ctemp);
}


void inipcaldatv3(pcaltype_tv3 *pc)  // initialize PCAL detection Type Ver 3.0
{
	int m;
    pc->ns=0;
    for(m=0; m< pc->chsize; m++){
        pc->rpcal[m]=0.0;
        pc->ipcal[m]=0.0;
        pc->sum1[m]=0.0;
        pc->sum2[m]=0.0;
    }
    for(m=0; m < pc->chsize; m++){
        pc->pcalf[m]=10.0e3;  // PCAL frequency 10kHz (default)
	}
    pc->kpp=0;
    //pc->mode=0;
    //pc->tmode=0;
}


void pcaldatnormv3(pcaltype_tv3 *pc)  // PCAL detection normalize (modified on 2003-07-24) 
{
	int m;
    double sum1,power;
    if(pc->mode!=0){
	    //for(k=0; k<= pc->kpp; k++){
        if(pc->ns!=0){
             for(m=0; m<pc->numch; m++){
                  sum1=pc->sum1[m];
                  power=pc->sum2[m]-(sum1*sum1)/(double)pc->ns;
                  pc->rpcal[m]/=sqrt(power);
                  pc->ipcal[m]/=sqrt(power);
              }
        } else {
             for(m=0; m<pc->numch; m++){
                  pc->rpcal[m]=0.0;
                  pc->ipcal[m]=0.0;
             }
        }
    } else {
        for(m=0; m<pc->numch; m++){
            pc->rpcal[m]=0.0;
            pc->ipcal[m]=0.0;
        }

    }    
}

void pcaldatnorm_1bit(pcaltype_tv3 *pc)  // PCAL detection normalize for 1 bit mode 2003-08-05 
{
	int m;
    double sum1;
    if(pc->mode!=0){
	    //for(k=0; k<= pc->kpp; k++){
        if(pc->ns!=0){
             for(m=0; m<pc->numch; m++){
                  sum1=pc->sum1[m];
                  pc->rpcal[m]=(2.0*pc->rpcal[m]-sum1)/sum1;
                  //pc->ipcal[m]=-(2.0*pc->ipcal[m]-sum1)/sum1;   
                  // signe change should not do here
                  // it will be done later in out_for_pvwave_1bit function
                  pc->ipcal[m]=(2.0*pc->ipcal[m]-sum1)/sum1;  
              }
        } else {
             for(m=0; m<pc->numch; m++){
                  pc->rpcal[m]=0.0;
                  pc->ipcal[m]=0.0;
             }
        }
    } else {
        for(m=0; m<pc->numch; m++){
            pc->rpcal[m]=0.0;
            pc->ipcal[m]=0.0;
        }

    }    
}

int checkheader(ip_vlbi_file_t *S) //renewed
{
/*
     2006-01-03

    Read 1st header to get sampling prameters and 
    to check sampler type either K5/VSSP or K5/VSSP32
    
       if first block is not header block, found header block automatically
       return value > 0 means above condition occurred
                    =-1 means EOF detection (file closed)

*/
      
       int iok, i, n, i32;
       int cdat[2]={1,4};
       int sdat[4]={1,2,4,8};
       int fdat[16]={40,100,200,500,1000,2000,4000,8000,16000, 
                     32000,64000,128000,256000,512000,1024000,2048000}; 
       union HEAD_VSSP hd;
       //fpos_t save;
       i32=0;
       iok=-1;

       S->vflag_old=S->vflag;                  // keep current valid flag
       S->dtime_vflag_old=S->dtime_vflag;  //keep time when vflag decision made
       S->vflag=0;

       //S->hd.format=0; // must be initialize before 1st time call!!

       hd.bbuf=S->hd.bhead;

       memset(S->hd.bhead, 0, 32);  // clear S->hd.head32 array with 0
       /* read 32bit data 1st*/
       S->bytesread = fread( S->hd.bhead, sizeof(char), 4, S->stream );
       if( ferror( S->stream ) || S->bytesread==0 ){
           if(S->bytesread==0){
               printf("checkheader: File EOF! (%s)\n",S->fname);
           } else {
               printf("checkheader: File Read Error! (%s)\n",S->fname);
           }
	       i=fclose( S->stream ) ;
           S->open_flag=0;
		   i32=-1;
           goto exit_end;
	   }

loop1:
       /* read 32 bit data and check sync patter by step-by-step */
       S->bytesread = fread( S->hd.bhead+4, sizeof( char ), 4, S->stream );
       if( ferror( S->stream ) || S->bytesread==0 ){
           if(S->bytesread==0){
               printf("checkheader: File EOF! (%s)\n",S->fname);
           } else {
               printf("checkheader: File Read Error! (%s)\n",S->fname);
           }
	       i=fclose( S->stream ) ;
           S->open_flag=0;
		   i32=-1;
           goto exit_end;
	   }
       /*  debug */
       /*
       printf("%02X%02X%02X%02X %02X\n",
           S->hd.bhead[0],S->hd.bhead[1],S->hd.bhead[2],S->hd.bhead[3],S->hd.bhead[7]);
       printf("%02X%02X%02X%02X %02X\n",
           hd.bbuf[0],hd.bbuf[1],hd.bbuf[2],hd.bbuf[3],hd.bbuf[7]);
       printf("%08X %02X\n",hd.hd->syncptn1,hd.hd->syncptn2);
       printf("totalsec0 %d %d\n",hd.hd->totalsec0, S->hd.bhead[5]*256+S->hd.bhead[4]);
       printf("AD SAMPLING %d %d %02X\n",hd.hd->bitwidth,hd.hd->fsample, S->hd.bhead[6]);
       exit(0);
       */
       /* check sync pattern */
       if(S->hd.format==0){
          if(hd.hd->syncptn1 == 0xFFFFFFFF && (unsigned char)hd.hd->syncptn2 == 0x8B) S->hd.format=1;// K5/VSSP
          if(hd.hd->syncptn1 == 0xFFFFFFFF && (unsigned char)hd.hd->syncptn2 == 0x8C) S->hd.format=2;// K5/VSSP32
          if(S->hd.format !=0){
             iok=0;  // sync detected
             S->hd.psync2=(unsigned char)hd.hd->syncptn2; // save second sync pattern
          }
       } else {
          if(hd.hd->syncptn1 == 0xFFFFFFFF && (unsigned char)hd.hd->syncptn2 == S->hd.psync2) iok=0;
       }
       if(iok==0){
           S->hd.sfreq=fdat[hd.hd->fsample];
           S->hd.adbit=sdat[hd.hd->bitwidth];
           S->hd.numch=cdat[hd.hd->chmode];
           S->hd.seconds=hd.hd->totalsec1*65536+hd.hd->totalsec0;
           sec2hms(S->hd.seconds,&S->hd.hh,&S->hd.mm,&S->hd.ss);
           if(S->hd.format == 2){ // more header read
                /* 1st read 4 byte to get AUX field size info */
                S->bytesread = fread( S->hd.bhead+8, sizeof( char ), 4, S->stream );
                if( ferror( S->stream ) || S->bytesread==0 ){
                    if(S->bytesread==0){
                        printf("checkheader: File EOF! (%s)\n",S->fname);
                    } else {
                        printf("checkheader: File Read Error! (%s)\n",S->fname);
                    }
	                i=fclose( S->stream ) ;
                    S->open_flag=0;
		            i32=-1;
                    goto exit_end;
                }
                S->hd.aux_size=hd.hd->auxsize;
                S->hd.year=hd.hd->year+2000;
                S->hd.totalday=hd.hd->totalday;
                S->hd.v_major=hd.hd->v_major;
                S->hd.v_minor=hd.hd->v_minor;
                //printf("debug1 %d %d %d\n",S->hd.aux_size,S->hd.year,S->hd.totalday);
                //printf("debug2 %d %d %d\n",S->hd.v_major,S->hd.v_minor,S->bytesread);

                /* read AUX field */
                S->bytesread = fread( S->hd.bhead+12, sizeof( char ), S->hd.aux_size, S->stream );
                if( ferror( S->stream ) || S->bytesread==0 ){
                    if(S->bytesread==0){
                        printf("checkheader: File EOF! (%s)\n",S->fname);
                    } else {
                        printf("checkheader: File Read Error! (%s)\n",S->fname);
                    }
	                i=fclose( S->stream ) ;
                    S->open_flag=0;
		            i32=-1;
                    goto exit_end;
                }
                /* parse AUX field data */
                S->hd.aux_fmt=hd.hd->fmt;
                S->hd.filter =hd.hd->filter;
                if(S->hd.aux_fmt==1 || S->hd.aux_fmt==2){
                    for(n=0; n<8; n++) S->hd.aux_hostname[n]=hd.hd->hostname[n];
                    S->hd.aux_hostname[8]='\0';
                    if(S->hd.aux_fmt==1){
                       for(n=0; n<8; n++) S->hd.aux_station[n]=hd.hd->station[n];
                       S->hd.aux_station[8]='\0';
                       S->hd.aux_sid2[0]=hd.hd->sid2[0]; S->hd.aux_sid2[1]=hd.hd->sid2[1];
                       S->hd.aux_sid2[2]='\0';
                    } else {
                       S->hd.aux_station[0]='\0';
                       S->hd.aux_sid2[0]='\0';
                    }
                } else {
                    S->hd.aux_hostname[0]='\0';
                    S->hd.aux_station[0]='\0';
                    S->hd.aux_sid2[0]='\0';
                }
                //printf("debug3 fmt,filter %d %d\n",S->hd.aux_fmt,S->hd.filter);
                //printf("debug4 host station sid2 (%s) (%s) (%s)\n",S->hd.aux_hostname,S->hd.aux_station,S->hd.aux_sid2);
               
           }
           goto exit_loop;
       } 
       if(i32 ==0) {
           printf("checkheader: First 64bit is not header block!\n");
           printf("checkheader:   Data file name (%s)\n",S->fname);
           printf("checkheader:   searching header.....");
       }
       /* read next 32 bit data */
       for (n=0; n<4; n++) S->hd.bhead[n]=S->hd.bhead[n+4];
       i32++;
       goto loop1;

exit_loop:

       if(i32 <0) return i32;
       if (iok == 0){   // sync detection
          if(i32 !=0) printf("  found after %d bits read!\n",i32*32);
          if(S->hd.format==1)
             printf("checkheader: Header (K5/VSSP) Sync Detected!!\n");
          if(S->hd.format==2)
             printf("checkheader: Header (K5/VSSP32) Sync Detected!!\n");
		  printf("checkheader:   File : %s\n",S->fname);
          printf("checkheader:   A/D(bits) %d  CHs %d  SFreq(kHz) %d",
                    S->hd.adbit,S->hd.numch, S->hd.sfreq);
          printf("  Time %02d:%02d:%02d  sec %d\n",
                    S->hd.hh,S->hd.mm,S->hd.ss,S->hd.seconds);
          // following 4 lines are just for keep backword compatibility
          // should be obsoleted 2006-01-03
          S->adbit=S->hd.adbit;
          S->numch=S->hd.numch;
          S->sfreq=S->hd.sfreq;
          S->hh=S->hd.hh; S->mm=S->hd.mm; S->ss=S->hd.ss; S->seconds=S->hd.seconds;
	   }
exit_end:
       if(i32 == 0) S->vflag=1;
       S->dtime_vflag=S->hd.seconds;
       //printf("checkheader:   vflag     dtime_vflag     %d %d\n",S->vflag,S->dtime_vflag);
       //printf("checkheader:   vflag_old dtime_vflag_old %d %d\n",S->vflag_old,S->dtime_vflag_old);
       return i32;
}

int DatRead1bitSv(ip_vlbi_file_t *S, 
                 unsigned char* ibuf,
		         unsigned char *idata, 
                 int numb, 
                 pcaltype_tv3 *pc, int *cpat, char *pcalcT)
{
    /* data read with PCAL detection for 1bit sampling data    
                  2004-08-28   use Takashima's algorythm for re-arranging 
                               data stream
				  2007-10-24   support all PCAL frequency with considerably
                               high speed
       
       S     --  IP VLBI File type variable
       ibuf  --  Buffer just read the data
       idata --  Data decoded to serial channel data (RETURN)
       numb  --  number of bytes to read  (usually usampl*numch/8)
                   numb should be multiplier of 4
       pc    --  pcal type variable    
       pc->mode;  =1;  all pcal frequency is same as ch1
                  =2;  pcal frequency is different
                  =0;  skip pcal detection
       cpat  -- correlation table for PCAL detection
       pcalcT -- correlation table of 8bit by 8bit


    */

	unsigned char a,b;
    //unsigned char iw1,iw2,iw3,iw4;
    char fname2[80];
    int i,j,k,m,ib, iunit;
    int fpcalbyte,fcal4,csum,ssum;
    //double pcalf=10.0e3;   // pcal frequency for debug purpose
    double pcalfold;
    double t,phai;
    int bcos,bsin;
    int iphai;
    int stab[]={1,1,0,0,1};
    int ctab[]={1,0,0,1,1};

    unsigned long *Lbuf; 
    unsigned long tbuf; 

    if(S->icnt == S->imax-1) { // header read
         if(checkheader(&(*S)) < 0) return -1;   //header read error
         if(pc->mode!=0 && pc->tmode==0){        // every 1 sec PCAL integration mode
            pc->kpp++;                           // PCAL pp# count up
            printf("DatRead1bitSv: pckpp, pcns %d %d\n",pc->kpp,pc->ns);// debug
         }
		 S->icnt=-1;       // to make first unit number counter 0
	}


    /* usampl x  channel  data read */

    if(S->numch==1){
        S->bytesread = fread( idata, sizeof( char ), numb, S->stream );
    } else {
        S->bytesread = fread( ibuf, sizeof( char ), numb, S->stream );
    }

    if(S->bytesread==0 || S->bytesread < (unsigned int)numb) {//EOF or near end of file     read next candidate file
        printf("DatRead1bitSv: EOF found. Try extended file if existed\n");
        i=fclose(S->stream);  
        S->open_flag=0;  
		// continued file read
        make_extended_filename(S->fname, fname2);  // file name updated
        printf("DatRead1bitSv: File to be read is %s\n",fname2);
        if( (S->stream  = fopen( fname2, "rb" )) == NULL ){
            printf("DatRead1bitSv: Extended file not found! \n");
            // to keep consistency for older version
            // closed file reopen to close at calling side
            S->stream = fopen( S->fname, "rb" );
            return -99;
        }
        S->open_flag=1; 
        printf("DatRead1bitSv: File found and opened\n");
        // file name update
        strcpy(S->fname,fname2);
        // rest of data read
        k=S->bytesread;
        i=numb-k;
        m = fread( ibuf+k, sizeof( char ), i, S->stream );
        S->bytesread=k+m;
    }


    if( ferror( S->stream ) ) return -2;    // read error
    if (feof( S->stream ) !=0) return -99;  // already reach EOF
	S->icnt++;         // read unit# counter count up.  First becomes 0

    /* parse K5 data to each channel serial data  */
    // 1bit sampling data only (must be guaranteed)

    if(S->adbit != 1) printf("Warning!! This is not 1bit sampling data\n");

   
    iunit=numb/S->numch;  // number of bytes per channel
    if(S->numch !=1){
        //iunit=numb/4;
        /*  start of Takashima's Algorythm  */
        for(ib=0; ib<numb/4; ib++){
            Lbuf=(unsigned long *)(ibuf+4*ib);  //NG for non-intel cpu such as alpha
            // original bit stream  *Lbuf (ch1:a ch2:b ch3:c ch4:d) 4bytes
            //8d 8c 8b 8a 7d 7c 7b 7a ..... 4d 4c 4b 4a 3d 3c 3b 3a 2d 2c 2b 2a 1d 1c 1b 1a

            tbuf = *Lbuf & 0xa5a5a5a5;
            tbuf += ((*Lbuf >> 3) & 0x0a0a0a0a) + ((*Lbuf << 3) & 0x50505050);

            // (*Lbuf & 0xa5a5a5a5)
            //8d  0 8b  0  0 7c  0 7a ..... 4d  0 4b  0  0 3c  0 3a 2d  0 2b  0  0 1c 0 1a
            // (*Lbuf >> 3) & 0x0a0a0a0a
            // 0  0  0  0 8c  0 8a  0  0  0  0 .....  0 4c  0 4a  0  0  0  0  0 2c  0 2a  0
            // (*Lbuf << 3) & 0x50505050
            // 0 7d  0 7b  0  0  0  0 .....  0 3d  0 3b  0  0  0  0  0 1d  0 1b  0  0 0  0
            // tbuf
            //8d 7d 8b 7b 8c 7c 8a 7a ..... 4d 3d 4b 3b 4c 3c 4a 3a 2d 1d 2b 1b 2c 1c 2a 1a

            tbuf = (tbuf & 0xcc33cc33) + ((tbuf >> 6) & 0x00cc00cc) + ((tbuf << 6) & 0x33003300);

            // (tbuf & 0xcc33cc33)
            //8d 7d  0  0 8c 7c  0  0 ..... 4d 3d  0  0 4c 3c  0  0  0  0 2b 1b  0  0 2a 1a
            // (tbuf >> 6) & 0x00cc00cc
            // 0  0  0  0  0  0  0  0 8b 7b .....  0  0  0  0  0  0 4b 3b  0  0 4a 3a 0  0
            // (tbuf << 6) & 0x33003300
            // 0  0 6d 5d  0  0 6c 5c .....  0  0 2d 1d  0  0 2c 1c  0  0  0  0  0  0 0  0
            // tbuf
            //8d 7d 6d 5d 8c 7c 6c 5c ..... 4d 3d 2d 1d 4c 3c 2c 1c 4b 3b 2b 1b 4a 3a 2a 1a

            tbuf = (tbuf & 0xf0f00f0f) + ((tbuf >> 12) & 0x0000f0f0) + ((tbuf << 12) & 0x0f0f0000);

            // (tbuf & 0xf0f00f0f)
            //8d 7d 6d 5d  0  0  0  0 .....  0  0  0  0 4c 3c 2c 1c  0  0  0  0 4a 3a 2a 1a
            // (tbuf >> 12) & 0x0000f0f0
            // 0  0  0  0  0  0  0  0 8b 7b ..... 6c 5c  0  0  0  0 8a 7a 6a 5a  0  0 0  0
            // (tbuf << 12) & 0x0f0f0000
            // 0  0  0  0 4d 3d 2d 1d  ..... 0  0  0  0  0  0  0  0  0  0  0  0  0  0 0  0
            // tbuf
            //8d 7d 6d 5d 4d 3d 2d 1d ..... 8c 7c 6c 5c 4c 3c 2c 1c 8a 7a 6a 5a 4a 3a 2a 1a

            // finally, this byte stream "tbuf" is |ch4|ch2|ch3|ch1|.
            idata[ib]   = (unsigned char) tbuf;   //ch1
            idata[ib+iunit]  = (unsigned char)(tbuf >> 16);  //ch2
            idata[ib+iunit*2]  = (unsigned char)(tbuf >> 8);  //ch3
            idata[ib+iunit*3]  = (unsigned char)(tbuf >> 24);  //ch4 
        }
        /*  End of Takashima's Algorythm */
   }

    // parseADdata_3(ibuf,numb,S->adbit,idata, &(*kpos));
	S->dtime=(double)S->seconds+S->unitp*(double)S->icnt;
	//printf("Dat1unitRead: numb, *kpos %d %d\n",numb,*kpos); //debug
    /* debug */
/*
	printf("S->dtime  %f\n",S->dtime);  //debug
    i=0;
    for(k=0; k<iunit; k++){
        a=idata[k+3*iunit]; // ch4 pick up
        //a=idata[k]; // ch1 pick up
        for(m=0; m<8; m++){
            if( (a & 1) ==0) {
                printf(" 0");
            } else {
                printf(" 1");
            }
            a = a >> 1;
            i++;
            if(i == 20) {
              printf("\n");
              i=0;
            }
        }
    }
    exit(0);
*/
    /* debug end */

    if(pc->mode ==0) return S->bytesread;

    /* PCAL detection */ 

    if(pc->mode==1){   // PCAL detection using all same frequency
        fpcalbyte=(int)(S->fsample/pc->pcalf[0])/8  ;  
        if ( (fpcalbyte % 4) !=0 ) goto method2;  // bit parse necessary
        fcal4=fpcalbyte/4 ;
	    
        for(m=0; m<S->numch; m++){
	       csum=0;
	       ssum=0;
           for (i=0; i<iunit; i++){
              //j=((S->icnt-1)*iunit + i)/fcal4 ;  // S->icnt start with 0, so -1 is not necessary
              j=(S->icnt*iunit + i)/fcal4 ; 
              a=idata[m*iunit+i];
              k=j % 4;
		      /* sin count */
              if( k ==0  || k == 1) { //  0-180 deg
		         ssum += cpat[a];
              } else if( k ==2  || k == 3) { //  180-360 deg
		         ssum += cpat[(unsigned char)~a];
              }

		      /*  cos count */

              if( k ==0  || k == 3) { //  0-90 or 270-360 deg
		          csum += cpat[a];
              } else if( k ==1  || k == 2){  //  90-270 deg
		          csum += cpat[(unsigned char)~a];
              }
           }
           pc->rpcal[m]+=(double)csum;
           pc->ipcal[m]+=(double)ssum;
           pc->sum1[m]+=(double)(iunit*8);
           pc->sum2[m]+=(double)(iunit*8); // not necessary now
        }
        pc->ns+=(iunit*8);       // processing # of samples update
        goto exitend;

 method2:;
	    // PCAL detection for 4MHz sampling 10kHz case
        // fpcalbyte = 50 is supposed
        //printf("Method 2 %d\n",fpcalbyte);
        for(m=0; m<S->numch; m++){
	       csum=0;
	       ssum=0;
           for (i=0; i<iunit; i++){
              //j=((icnt-1)*nbytes + i)/fcal4 ;
              k=(S->icnt*iunit + i) % fpcalbyte ;
              a=idata[m*iunit+i];
		      /* sin count */
              if( k >=0  && k < 25) { //  0-180 deg
		         ssum += cpat[a];
              } else {//  180-360 deg
		         ssum += cpat[(unsigned char)~a];
              }

		      /*  cos count */

              if( (k >=0  && k < 12) || (k > 37 && k< 50)) { //  0-90 or 270-360 deg
		         csum += cpat[a];
              } else if( k > 12 && k < 37){  //  90-270 deg
		         csum += cpat[(unsigned char)~a];
              } else if (k==12){   // 90 deg
                 b=a & 0xF0;
                 b = b >> 4;
                 for(j=0; j<4; j++){
                    if( (b & 1) == 1) csum++;
                    b= b>>1;
                 }
                 b= (~a) & 0xF;
                 for(j=0; j<4; j++){
                    if( (b & 1) == 1) csum++;
                    b= b>>1;
                 }
              } else if (k==37) {  // 270 deg
                 b=(~a) & 0xF0;
                 b = b >> 4;
                 for(j=0; j<4; j++){
                    if( (b & 1) == 1) csum++;
                    b= b>>1;
                 }
                 b= a & 0xF;
                 for(j=0; j<4; j++){
                    if( (b & 1) == 1) csum++;
                    b= b>>1;
                 }
              }
           }
           pc->rpcal[m]+=(double)csum;
           pc->ipcal[m]+=(double)ssum;
           pc->sum1[m]+=(double)(iunit*8);
           pc->sum2[m]+=(double)(iunit*8);
        }
        pc->ns+=(iunit*8);       // processing # of samples update

    } else {   // for pc->mode = 2

        for(k=0; k< iunit; k++){   // byte unit
            pcalfold=0.0;
            for(m=0; m<S->numch; m++){
                a=idata[m*iunit+k];
                // 1byte PCAL detection pattern generate
                if(pc->pcalf[m] != pcalfold){
                    bcos=0;
                    bsin=0;
                    for(i=0; i<8; i++){    // bit unit
                       t=S->dtime+(double)(k*8+i)*S->tsample;
                       phai=fmod(TWOPI*pc->pcalf[m]*t,TWOPI)*PIDE;
                       iphai=(int)(phai/90.0);
                       //bcos=(bcos << 1) | (ctab[iphai] & 0x1);
                       //bsin=(bsin << 1) | (stab[iphai] & 0x1);
                       bcos=(bcos >> 1) | ((ctab[iphai] & 0x1) << 7);
                       bsin=(bsin >> 1) | ((stab[iphai] & 0x1) << 7);
                    }
                    pcalfold=pc->pcalf[m];
                }
                //printf("m,pcalf,bcos,bsin %d %f %d %d\n",m,pc->pcalf[m],bcos,bsin);//debug

                // PCAL correlation
                csum=(int)pcalcT[a*256+bcos];
                ssum=(int)pcalcT[a*256+bsin];
                pc->rpcal[m]+=(double)csum;
                pc->ipcal[m]+=(double)ssum;
                pc->sum1[m]+=(double)8;
                pc->sum2[m]+=(double)8; // not necessary now

            }
            pc->ns+=8;       // processing # of samples update
        }

    }
exitend:

	return S->bytesread;
}



long int Dat1unitRead_w_Pcdetv3(ip_vlbi_file_t *S, unsigned char* ibuf,
					  int *idata, int numb, int *kpos, pcaltype_tv3 *pc)
{
    /* data read with PCAL detection    Version 3.0   2003-02-11
                                                      2003-08-02
    
       pc->mode;  =1;  all pcal frequency is same as ch1
                  =2;  pcal frequency is different
                  =0;  skip pcal detection
    */

    int k,m,i;
    char fname2[80];
    double t,fs,fc,phai,w1,w2;
    double pcalf=10.0e3;   // pcal frequency for debug purpose
	//printf("S->icnt numb %d %d\n",S->icnt,numb);  // debug
    if(S->icnt == S->imax-1) { // header read
         if(checkheader(&(*S)) < 0) return -1;   //header read error
         if(pc->mode!=0 && pc->tmode==0){        // every 1 sec PCAL integration mode
            pc->kpp++;                           // PCAL pp# count up
            printf("Dat1unitRead_w_Pcdetv3: pckpp, pcns %d %d\n",pc->kpp,pc->ns);// debug
         }
		 S->icnt=-1;       // to make first unit number counter 0
	}

    /* usampl x  channel  data read */

    // S->bytesread = fread( S->ibuf, sizeof( char ), S->numb, S->stream );
    S->bytesread = fread( ibuf, sizeof( char ), numb, S->stream );
    //printf("bytes read 2 %d\n",S->bytesread); //debug

    // extended file read process added on 2003-08-02
    if(S->bytesread==0 || S->bytesread < (unsigned int)numb) {//EOF or near end of file     read next candidate file
        printf("Dat1unitRead_w_Pcdetv3: EOF found. Try extended file if existed\n");
        i=fclose(S->stream); // 
        S->open_flag=0; //2005-11-08
        make_extended_filename(S->fname, fname2);  // file name updated
        printf("Dat1unitRead_w_Pcdetv3: File to be read is %s\n",fname2);
        if( (S->stream  = fopen( fname2, "rb" )) == NULL ){
            printf("Dat1unitRead_w_Pcdetv3: Extended file not found! \n");
            // to keep consistency for older version
            // closed file reopen to close at calling side
            S->stream = fopen( S->fname, "rb" );
            return -99;
        }
        S->open_flag=1; //2005-11-08
        printf("Dat1unitRead_w_Pcdetv3: File found and opened\n");
        // file name update
        strcpy(S->fname,fname2);
        // rest of data read
        k=S->bytesread;
        i=numb-k;
        m = fread( ibuf+k, sizeof( char ), i, S->stream );
        S->bytesread=k+m;
    }


    if( ferror( S->stream ) ) return -2;    // read error
    if (feof( S->stream ) !=0) return -99;  // already reach EOF
	S->icnt++;         // read unit# counter count up.  First becomes 0

    /* parse each channel data */

    //parseADdata_3(S->ibuf,S->numb,S->adbit,idata, &(*kpos));
    parseADdata_3(ibuf,numb,S->adbit,idata, &(*kpos));
    //parseK5data(ibuf,numb,S->adbit,idata, &(*kpos)); <=This doesn't show any speed up
	S->dtime=(double)S->seconds+S->unitp*(double)S->icnt;
	//printf("Dat1unitRead: numb, *kpos %d %d\n",numb,*kpos); //debug
	//printf("S->dtime  %f\n",S->dtime);  //debug

    if(pc->mode ==0) return S->bytesread;

    /* PCAL detection */ 

    for(k=0; k<*kpos/S->numch; k++){
        t=S->dtime+(double)k*S->tsample;
        if(pc->mode==1){
            phai=fmod(TWOPI*pc->pcalf[0]*t,TWOPI);
            fc=cos(phai);
            fs=sin(phai);
            // following is debug simulating 1bit approximation
            //if(fc >= 0.0) fc=1.0;
            //if(fc < 0.0) fc=-1.0;
            //if(fs >= 0.0) fs=1.0;
            //if(fs < 0.0) fs=-1.0;
        }
        //printf("dtime, t phai fc fs %g %g %g %g %g\n",S->dtime,t,phai,fc,fs);
        //printf("dtime t %f %f %g\n",S->dtime,t, S->tsample);
        for(m=0; m<S->numch; m++){
    //printf("pcalmode %d %f\n",pc->mode,pc->pcalf[m]);
            if(pc->mode==2){
                phai=fmod(TWOPI*pc->pcalf[m]*t,TWOPI);
                fc=cos(phai);
                fs=sin(phai);
                // following is debug simulating 1bit approximation
                //if(fc >= 0.0) fc=1.0;
                //if(fc < 0.0) fc=-1.0;
                //if(fs >= 0.0) fs=1.0;
                //if(fs < 0.0) fs=-1.0;
            }
            w1=(double)idata[k*S->numch+m];
            w2=w1*w1;
            pc->rpcal[m]+=fc*w1;
            pc->ipcal[m]+=fs*w1;
            pc->sum1[m]+=w1;
            pc->sum2[m]+=w2;
        }
        pc->ns++;       // processing # of samples update
    }
	return S->bytesread;
}

void make_extended_filename(char *fname, char *fname2)
// file nmae updated as
//     XXXXXX.dat -> XXXXXX.dat1 -> XXXXXX.dat2 .....
//  every call
{
    char abuf[80],cbuf[10];
    int n,len;
    len=strlen(fname);
    // get last "." position
    for(n=len-1; n >=0; n--){
        if(fname[n]=='.') break;
    }
    strncpy(abuf,fname,n);
    abuf[n]='\0';
    // now abuf contains a portion before last '.'
    strcpy(cbuf,fname+n+1);
    // cbuf contains a portion after last '.'
    // check cbuf accompanied by any number
    n=0;
    len=strlen(cbuf);
    if(len > 3){ // any number is accompanied 
        n=atoi(cbuf+3);
    }
    n++;
    sprintf(fname2,"%s.dat%d\0",abuf,n);
}



/******** copy from libcor end ************/


typedef struct pcal_plot    /* parameter structure for pcalplot */
{            
    char g_device[30]; // current graphic device name 
    char g_display[30]; // graphic device name for display in case of GNUPLOT
    char g_ps_device[30]; // graphic device name for POSTSCRIPT
    //char pgdevice[30];    // pointer to PGPLOT device name  
    char fname[128];       // data file name
    char pfile[128];    // filename containing detected PCAL phase and amp data 
    char ps_file[128];    // POSTSCRIPT filename
    double pcalf[4];     //  pcal frequency table
    double rpcal[4],ipcal[4]; // rpcalipcal
    int numch;        
    float tmin,tmax;
    float phsmin,phsmax;
    float ampmin,ampmax;
    double stime; // start time (seconds past 0h UTC)
    double dtime; // current time (seconds past 0h UTC)
    double scanlength; // scan length (sec)
    double tinteg; // tinteg
    int hh; // start time hour
    int mm; // start time minute
    int ss;  // start time second
    vssp_head_t hd;   //VSSP header info
    // belows are for GNUPLOT 
    FILE *gp;   // for GNUPLOT 
    char htitl1[256], xtit1[60], ytit1[60];
    char htitl2[256], xtit2[60], ytit2[60];
} pcal_plot_t;



typedef struct run_param_pcalcheck    /* run parameter structure for pcalcheck */
{            
    char fname[128];       // file name
    char pfile[128];       // pcal detection results file name
    int startoffset;      // processing start time offset in sec
    double pcalf[4];      // pcal frequency by CH (Hz)
    double sekibun;       // given integration time
    double coffset;       // clock offset
    double roffset;       // clock rate
    double sfmhz;
    int adbit;
    int numch;
    int vssp32;           // 1: VSSP32 mode 
    float t1pp;           // given 1pp period (sec)
    int header_no_check;  // 1: header no check
    int after;            // after plot flag 1: after plot 0: real time plot
    char pgdevice[20];    // PGPLOT device name  
    char ps_file[128];    // POSTSCRIPT file name  
	char program_name[60];  
} run_param_pcalcheck_t;


void run_param_get_4_pcalcheck(int argc, char* argv[], run_param_pcalcheck_t *runp)
/*
   get pcalcheck run parameters   2007-11-03
*/
{
    char cbuf[40];
    int n;
    runp->header_no_check=0;
    runp->vssp32=0;
    runp->sekibun=0.0;
    /* set graphic device */
#ifdef USE_GNUPLOT   // using gnuplot
    runp->after=1;
    strcpy(runp->pgdevice,G_DISPLAY);
#else
    runp->after=0;
    strcpy(runp->pgdevice,select_env_dflt(PGDISP, PGDISP_DFLT));
#endif
    strcpy(runp->ps_file,PS_OUT_FILE);

    for(n=0; n<4; n++) runp->pcalf[n]=10.0e3;

	if (argc>=2){
          strcpy(runp->fname,*(argv+1));
	}
    if(cmnd_get(argc,argv,"-nocheck",cbuf)==0)    runp->header_no_check=1; //header no check mode
    if(cmnd_get(argc,argv,"-sfreq",cbuf)==0)    runp->sfmhz=atof(cbuf); //sampling freq (MHz)
    if(cmnd_get(argc,argv,"-adbit",cbuf)==0)    runp->adbit=atoi(cbuf); //AD bits
    if(cmnd_get(argc,argv,"-numch",cbuf)==0)    runp->numch=atoi(cbuf); //# of channel
    if(cmnd_get(argc,argv,"-fa",cbuf)==0){//PCAL frequency (Hz)
        for(n=0; n<4; n++) runp->pcalf[n]=1000.0*atof(cbuf);
    }
    if(cmnd_get(argc,argv,"-integ",cbuf)==0)    runp->sekibun=atof(cbuf); //integration period
    if(cmnd_get(argc,argv,"-f1",cbuf)==0) runp->pcalf[0]=1000.0*atof(cbuf); //CH1 PCAL freq
    if(cmnd_get(argc,argv,"-f2",cbuf)==0) runp->pcalf[1]=1000.0*atof(cbuf); //CH1 PCAL freq
    if(cmnd_get(argc,argv,"-f3",cbuf)==0) runp->pcalf[2]=1000.0*atof(cbuf); //CH1 PCAL freq
    if(cmnd_get(argc,argv,"-f4",cbuf)==0) runp->pcalf[3]=1000.0*atof(cbuf); //CH1 PCAL freq
    if(cmnd_get(argc,argv,"-vssp32",cbuf)==0) runp->vssp32=1; //set VSSP32 mode
    if(cmnd_get(argc,argv,"-pg",cbuf)==0){
       strcpy(runp->pgdevice,cbuf); // PGPLOT device set 
    }
    if(cmnd_get(argc,argv,"-ps",cbuf)==0){
       strcpy(runp->pgdevice,G_PS_DEVICE); // postscript for PGPLOT device set 
    }
    if(cmnd_get(argc,argv,"-after",cbuf)==0)  runp->after=1; //plot after flag
    if(cmnd_get(argc,argv,"-noa",cbuf)==0)    runp->after=0; //plot after flag
    if(cmnd_get(argc,argv,"-out",cbuf)==0)  strcpy(runp->pfile,cbuf);

}

void print_usage(void)
{
    printf("pcalcheck (%s)\n",PROGRAM_VERSION);
#ifdef USE_GNUPLOT   
    printf("   compiled for GNUPLOT\n\n");
#else
    printf("   compiled for PGPLOT\n\n");
#endif
    printf("Usage: \n");
    printf("  pcalcheck k5file [option]\n");
    printf("   or\n");
    printf("  pcalcheck ofile [-pgplot device | -ps]  --- for re-plot\n");
    printf("\n");
    printf("    where  k5file ---- K5 file name\n");
    printf("           ofile  ---- pcalcheck output file name\n");
    printf("       [option]\n");
    printf("           -fall pcalf --- set PCAL frequency for all channels (kHz)\n");
    printf("                             (default 10kHz)\n");
    printf("           -f1 pcalf1 --- set PCAL frequency for CH1 (kHz)\n");
    printf("           -f2 pcalf2 --- set PCAL frequency for CH2 (kHz)\n");
    printf("           -f3 pcalf3 --- set PCAL frequency for CH3 (kHz)\n");
    printf("           -f4 pcalf4 --- set PCAL frequency for CH4 (kHz)\n");
    printf("           -integ integ --- set integration period (sec)\n");
    printf("                             (default is all scan period)\n");
#ifndef USE_GNUPLOT   
    printf("           -pgplot device -- set PGPLOT device\n");
    printf("                             /NULL for suppression\n");
    printf("                             /CPS for colored postscript\n");
    printf("                             (default is  display)\n");
#endif
    printf("           -ps  -- compulsory set graphic device to POSTSCRIPT\n");
    printf("                        (valid for re-plot mode)\n");
    printf("           -after   --- plot graph after all data processing\n");
    printf("                       (this is default in case of GNUPLOT)\n");
    printf("           -noafter --- plot and update graph with data processing\n");
    printf("                       (this is default in case of PGPLOT)\n");
    printf("           -out ofile -- set output file name\n");
    printf("                        (default is \"pcalcheck_out.txt\")\n");
    printf("           -nocheck --- no header check\n");
    printf("            <<following parameters for nocheck mode>>\n");
    printf("           -vssp32  --- set VSSP32 mode\n");
    printf("           -sfreq sfreq --- set sampling frequency (MHz)\n");
    printf("           -adbit adbit --- set AD bits\n");
    printf("           -numch numch --- set # of channels\n");
}


int64_t get_file_size(char *fname)
{
    /*  get file size opened */
	int result;
#ifdef WIN32                       //2006-09-20
    struct _stati64 info;    // for VC++    
    result=_stati64(fname,&info);  // for VC++
#else
    struct stat info;
    result=stat(fname,&info);
#endif
	if(result != 0) return (int64_t)(-1);
	else
    return (int64_t)info.st_size;
    // use %lld to print out whole number that has int64_t type
}


int parseline2item(char *line, double *data) 
{
/*
    parse one line data to each item seprated by space

  return
     # of items

*/
    char abuf[80];
    char *cset=" ";
    char *p=NULL;
    int k;
   
    k=0;
    sprintf(abuf,"%s\0",strtok(line, cset));
    data[k]=atof(abuf);   // 1st item
    k++;
    while((p=strtok(NULL,cset))){
        sprintf(abuf,"%s\0",p);
        data[k]=atof(abuf);
        k++;
    }
    return k;
}

void get_nth_item_in_a_line(char *line, int n, char *abuf) 
{
/*
    parse one line data and get nth item seprated by space

*/
    char cbuf[256];
    char *cset=" ";
    char *p=NULL;
    int k,len;

    strcpy(cbuf,line);
    /* remove LF code just in case*/
    len=strlen(cbuf);
    if(cbuf[len-1]==0x0A) cbuf[len-1]='\0';
   
    k=0;
    sprintf(abuf,"%s\0",strtok(cbuf, cset));
    k++;
    if(k==n) goto exithere;
    while((p=strtok(NULL,cset))){
        sprintf(abuf,"%s\0",p);
        k++;
        if(k==n) break;
    }
exithere:;
}

int pcal_info_file_check(char *fname)
{
/* 
   PCAL result file check 

   return
   
   -1: given file is not PCAL results file
    0: OK;
*/
    FILE *fo;
    char cbuf[256];
    int ierr;
    ierr=0;
    if((fo = fopen(fname, "r" )) == NULL ){
        printf("pcal_info_file_check: File (%s) Open Error! \n",fname);
        return -1;
    }
    /* check first line */
    if(fgets(cbuf, 256, fo) != NULL){
        if(strncmp(cbuf,"# PCAL",6)!=0) ierr=-1;  // this is not PCAL File
    }
    fclose(fo);
    return ierr;
}

int pcal_info_get(pcal_plot_t *plotP)
{
/* 
   read PCAL check out file and get information 

   return
   
   -1: given file is not PCAL results file
    0: OK;
*/

    FILE *fo;
    char cbuf[256];
    char abuf[128];
    char wbuf[10];
    int ierr;


    //printf("Debug after \n");
    ierr=0;
    /* initialize parameters */
    plotP->hd.aux_hostname[0]='\0';
    plotP->hd.aux_station[0]='\0';
    plotP->fname[0]='\0';
    plotP->hd.format=0;
    plotP->hd.year=0;
    plotP->hd.totalday=0;
    plotP->hh=0;
    plotP->mm=0;
    plotP->ss=0;
    plotP->stime=0.0;
    plotP->scanlength=0.0;
    plotP->hd.sfreq=0;
    plotP->hd.adbit=0;
    plotP->hd.numch=0;
    plotP->hd.filter=0;
    plotP->numch=0;
    plotP->pcalf[0]=0.0;
    plotP->pcalf[1]=0.0;
    plotP->pcalf[2]=0.0;
    plotP->pcalf[3]=0.0;

    if((fo = fopen(plotP->pfile, "r" )) == NULL ){
        printf("pcal_info_get: File (%s) Open Error! \n",
                plotP->pfile);
        ierr=-1;
		goto exithere;
    }
    /* check first line */
    if(fgets(cbuf, 256, fo) != NULL){
        if(strncmp(cbuf,"# PCAL",6)!=0){        // this is not PCAL File
           fclose(fo);
           ierr=-1;
           goto exithere;
        }
    }
    while(fgets(cbuf, 256, fo) != NULL){
     if(cbuf[0]!='#') continue;  // this is not comment line (nothing to do)
     if(strncmp(cbuf,"# File",6)==0){        // File name block
         get_nth_item_in_a_line(cbuf, 4, abuf);
         strcpy(plotP->fname,abuf);
     }
     if(strncmp(cbuf,"# Sampler Type",14)==0){   // Sampler type
         get_nth_item_in_a_line(cbuf, 5, abuf);
         if(strncmp(abuf,"K5/VSSP32",9)==0){   // K5/VSSP32
            plotP->hd.format=2;
         } else {
            plotP->hd.format=1;
         }
     }
     if(strncmp(cbuf,"# Start Time(UT)",14)==0){   // Start time
         if(plotP->hd.format==2){
             get_nth_item_in_a_line(cbuf, 5, abuf);  // year/date
             strncpy(wbuf,abuf,4); wbuf[4]='\0';
             plotP->hd.year=atoi(wbuf);
             strncpy(wbuf,abuf+5,3); wbuf[3]='\0';
             plotP->hd.totalday=atoi(wbuf);

             get_nth_item_in_a_line(cbuf, 6, abuf);  // hh:mm:ss
         } else {
             get_nth_item_in_a_line(cbuf, 5, abuf);  // hh:mm:ss
         }
         strncpy(wbuf,abuf,2); wbuf[2]='\0';
         plotP->hh=atoi(wbuf);
         strncpy(wbuf,abuf+3,2); wbuf[2]='\0';
         plotP->mm=atoi(wbuf);
         strncpy(wbuf,abuf+6,2); wbuf[2]='\0';
         plotP->ss=atoi(wbuf);
         plotP->stime=(double)(plotP->hh*3600+plotP->mm*60+plotP->ss);
     }
     if(strncmp(cbuf,"# Scan Length(sec)",18)==0){   // Scan length
         get_nth_item_in_a_line(cbuf, 5, abuf);  
         plotP->scanlength=atof(abuf);
     }
     if(strncmp(cbuf,"# Sampler Host PC",17)==0){   // Sampler host PC
         get_nth_item_in_a_line(cbuf, 6, abuf); 
         strcpy(plotP->hd.aux_hostname,abuf);
     }
     if(strncmp(cbuf,"# Sampler Station",17)==0){   // Sampler station
         get_nth_item_in_a_line(cbuf, 5, abuf);
         strcpy(plotP->hd.aux_station,abuf);
     }
     if(strncmp(cbuf,"# Fs=",5)==0){   // Sampler parameters
         get_nth_item_in_a_line(cbuf, 3, abuf);
         plotP->hd.sfreq=(int)(atof(abuf)*1000.0);
         get_nth_item_in_a_line(cbuf, 6, abuf);
         plotP->hd.adbit=atoi(abuf);
         get_nth_item_in_a_line(cbuf, 8, abuf);
         plotP->hd.numch=atoi(abuf);
         if(plotP->hd.format==2){
            get_nth_item_in_a_line(cbuf, 10, abuf);
            plotP->hd.filter=atoi(abuf);
         }
         plotP->numch=plotP->hd.numch;
     }
     if(strncmp(cbuf,"# CH #1",7)==0){   // Pcal frequency
         get_nth_item_in_a_line(cbuf, 4, abuf);
         plotP->pcalf[0]=atof(abuf)*1000.0;
         if(plotP->hd.numch==4){
            get_nth_item_in_a_line(cbuf, 8, abuf);
            plotP->pcalf[1]=atof(abuf)*1000.0;
            get_nth_item_in_a_line(cbuf, 12, abuf);
            plotP->pcalf[2]=atof(abuf)*1000.0;
            get_nth_item_in_a_line(cbuf, 16, abuf);
            plotP->pcalf[3]=atof(abuf)*1000.0;
         }
         break; // PCAL info is the last item
     }
    }
    fclose(fo);
exithere:;
    return ierr;
}

void pcal_plot_start(pcal_plot_t *plotP)
{
      char htitl1[256], xtit1[60], ytit1[60];
      char htitl2[256], xtit2[60], ytit2[60];
      char lab[128],lab1[256],lab2[256],lab3[128],fname[128];
      int n,len;
      int hh,mm,ss,seconds;
      double pcalf[4];
#ifdef USE_GNUPLOT   // using GNUPLOT
      //char *ftemp="pcal.tmp";
      char ftemp[][20]={"pcal1.tmp","pcal2.tmp","pcal3.tmp","pcal4.tmp"};
      FILE *fo;
      //int npmax;

	   plotP->gp= popen(GNUPLOT_EXE,"w");
       if(plotP->gp==NULL) goto exithere;
	   fprintf(plotP->gp,"set terminal %s\n",plotP->g_device);
	   fflush(plotP->gp);
	   fprintf(plotP->gp,"set multiplot\n");
	   fprintf(plotP->gp,"set size 1.0,0.5\n");

#else   // using PGPLOT
      if (cpgopen(plotP->g_device) < 1) goto exithere;
      if(strncmp(plotP->g_device,"/GW",3)==0) cpgeras();   // erase just in case 2007-10-26

      if(strncmp(plotP->g_device,"/XWIN",5)==0 ||
         strncmp(plotP->g_device,"/XSER",5)==0 ) { /* set background white 
                                                   and text black */
         cpgscr(0,1.0,1.0,1.0);  /* Background color */
         cpgscr(1,0.0,0.0,0.0);  /* Foreground color */
      }
      cpgsubp(1,2);  // divide plot pannel into 1X2
#endif

      // set title

      strcpy(fname,plotP->fname);
      len=strlen(fname);
      for(n=0; n<len; n++){  // convert '\' to '/' to avoid misoperation in plot
          if(fname[n]=='\\') fname[n]='/';
      }

      // set PCAL frequency in kHz
      for(n=0; n<plotP->numch; n++) pcalf[n]=plotP->pcalf[n]/1000.0;
      if(plotP->numch==1){
         sprintf(lab,"Ch#1: %.1fkHz\0",pcalf[0]);
      } else {
         sprintf(lab,"Ch#1: %.1fkHz  Ch#2: %.1fkHz  Ch#3: %.1fkHz  Ch#4: %.1fkHz\0",
          pcalf[0],pcalf[1],pcalf[2],pcalf[3]);
      }

      // set start time

      seconds=(int)plotP->stime;
      hh=seconds/3600;
      mm=(seconds % 3600)/60;
      ss=seconds % 60;

      if(plotP->hd.format==2){ // VSSP32 
         sprintf(lab1,"Start Time(UT) %04d/%03d %02d:%02d:%02d   \0",
             plotP->hd.year,plotP->hd.totalday,hh,mm,ss);
      } else {
         sprintf(lab1,"Start Time(UT) %02d:%02d:%02d   \0",
            hh,mm,ss);
      }
      plotP->hh=hh; plotP->mm=mm; plotP->ss=ss;

      strcat(lab1,lab);


      //set sampling parameters
      if(plotP->hd.format==2){ // VSSP32 
          sprintf(lab2,"VSSP32  Fs=%.1fMHz AD=%2d #_of_CH=%1d  LPF=%dMHz  \0",
              plotP->hd.sfreq/1000.0,plotP->hd.adbit,
              plotP->hd.numch,plotP->hd.filter);
      } else {
          sprintf(lab2,"VSSP  Fs=%.1fMHz AD=%2d #_of_CH=%1d  \0",
              plotP->hd.sfreq/1000.0,plotP->hd.adbit,
              plotP->hd.numch);
      }

      // set station name, host PC info

      lab3[0]='\0';
      if(plotP->hd.format==2){ // VSSP32 
         sprintf(lab3," Station: %s  Host PC: %s\0",
            plotP->hd.aux_station, plotP->hd.aux_hostname);
      }


      strcat(lab2,lab3);

//      cpgsubp(1,2);  // divide plot pannel into 1X2

      // set title axis label for PLOT#1
      sprintf(htitl1,"PCAL PHASE (%s)\0",fname);
      strcpy(xtit1,"TIME (sec)");
      strcpy(ytit1,"PHASE (deg)");

      // set title axis label for PLOT#2
      sprintf(htitl2,"PCAL AMPLITUDE (%s)\0",fname);
      strcpy(xtit2,"TIME (sec)");
      strcpy(ytit2,"AMPLITUDE");


      //for(n=0; n<4; n++){
      /* define coordinate range of phase graph */
      plotP->tmin=0.0;
      plotP->tmax=(float)plotP->scanlength;
      plotP->phsmin=-180.0;
      plotP->phsmax=180.0;
      /* define coordinate range of amp graph */
      plotP->ampmin=0.0;
      plotP->ampmax=(float)0.15;

#ifdef USE_GNUPLOT   // using GNUPLOT

      // dummy data prepare
      if( (fo  = fopen( ftemp[0],"w" )) == NULL ){
         printf(" File (%s) Open Error! \n",ftemp[0]);
         goto exithere;
      } else {
         //printf(" File (%s) Opened for GNUPLOT data write! \n",ftemp);
         fprintf(fo, "%g %g\n",plotP->tmin,plotP->phsmin);
      }
      fclose(fo);

      // PLOT#1
      fprintf(plotP->gp,"set origin 0.0,0.5\n");
	  fprintf(plotP->gp,"set xrange [%g:%g]\n",plotP->tmin, plotP->tmax);
	  fprintf(plotP->gp,"set yrange [%g:%g]\n",plotP->phsmin, plotP->phsmax);
      fprintf(plotP->gp,"set ytics -180, 90\n");
      
      /* label the axes */
	  fprintf(plotP->gp,"set title \"%s\"\n",htitl1);
	  fprintf(plotP->gp,"set xlabel \"%s\"\n",xtit1);
	  fprintf(plotP->gp,"set ylabel \"%s\"\n",ytit1);
	  fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.14 center\n",lab1);
	  fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.05 center\n",lab2);
	  fprintf(plotP->gp,"plot \"%s\" with dots notitle\n",ftemp[0]);


      // PLOT#2
	  fprintf(plotP->gp,"unset label\n");  // this is necessary otherwise 

      fprintf(plotP->gp,"set origin 0.0,0.0\n");
	  fprintf(plotP->gp,"set xrange [%g:%g]\n",plotP->tmin, plotP->tmax);
	  fprintf(plotP->gp,"set yrange [%g:%g]\n",plotP->ampmin, plotP->ampmax);
      fprintf(plotP->gp,"set ytics autofreq\n");
      
      /* label the axes */
	  fprintf(plotP->gp,"set title \"%s\"\n",htitl2);
	  fprintf(plotP->gp,"set xlabel \"%s\"\n",xtit2);
	  fprintf(plotP->gp,"set ylabel \"%s\"\n",ytit2);
	  fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.14 center\n",lab1);
	  fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.05 center\n",lab2);
	  fprintf(plotP->gp,"plot \"%s\" with dots notitle\n",ftemp[0]);
	  fprintf(plotP->gp,"unset label\n");  // this is necessary otherwise 
	  fflush(plotP->gp);
	  //getchar();  // this is necessary to keep display otherwise deleted

      strcpy(plotP->htitl1,htitl1);
      strcpy(plotP->xtit1,xtit1);
      strcpy(plotP->ytit1,ytit1);
      strcpy(plotP->htitl2,htitl2);
      strcpy(plotP->xtit2,xtit2);
      strcpy(plotP->ytit2,ytit2);

#else   // using PGPLOT

      cpgsch(2.0);  // set text height 2.0  (default is 1)
      cpgslw(2);    // set line width
      cpgsci(1);  /* black */
      //cpgenv(plotP->tmin, plotP->tmax, plotP->phsmin, plotP->phsmax, 0, 0);   // linear scale
      cpgenv(plotP->tmin, plotP->tmax, plotP->phsmin, plotP->phsmax, 0, -2);   // linear scale
      // cpgenv will move window automatically to next
      
      /* label the axes */
      cpgpanl(1,1);
      cpgbox("BCNST",0.0,0,"BCNST",90.0,3);  // to improve Y axis appearance
      //cbuf[0]='N';
      //cpgaxis(cbuf,plotP->tmin, plotP->phsmin,plotP->tmin, plotP->phsmax,
      //   plotP->phsmin,plotP->phsmax,90.0,3,0.0,1.0,0.5,-0.5,0.0);
      cpglab(xtit1,ytit1,htitl1);
      /* sub label write*/
      cpgsch(1.5);
      cpgmtxt("T",1.5,0.5,0.5,lab1);
      cpgmtxt("T",0.5,0.5,0.5,lab2);
      cpgsch(2.0);

      /* define coordinate range of amp graph */
//      plotP->ampmin=0.0;
//      plotP->ampmax=(float)0.15;
      /* label the axes */
      cpgenv(plotP->tmin, plotP->tmax, plotP->ampmin, plotP->ampmax, 0, 0);   // linear scale
      cpgpanl(1,2);
      //sprintf(htitl,"PCAL AMPLITUDE (%s)\0",fname);
      //strcpy(ytit,"AMPLITUDE");
      cpglab(xtit2,ytit2,htitl2);
      /* sub label write*/
      cpgsch(1.5);
      cpgmtxt("T",1.5,0.5,0.5,lab1);
      cpgmtxt("T",0.5,0.5,0.5,lab2);
      cpgsch(2.0);
#endif

exithere:;
}


void pcal_plot_end(pcal_plot_t *plotP)
{
/* end of all plot */
#ifdef USE_GNUPLOT   // using gnuplot

	   fflush(plotP->gp);
  #ifdef WIN32
       if(strncmp(plotP->g_device,"windows",7)==0){ // display
	     getchar();  // this is necessary to keep display otherwise deleted
       }
  #endif
       pclose(plotP->gp);
#else                // using pgplot

      if(strncmp(plotP->g_device,"/GW",3)!=0){
         cpgclos();
      }
#endif
}


void pcal_plot_point(pcal_plot_t *plotP)
{
      float tdat[2], ydat1[2], ydat2[2];
      double w1,w2,w3,w4;
      int n;
#ifdef USE_GNUPLOT   // using GNUPLOT
      char ftemp[][20]={"pcal1.tmp","pcal2.tmp","pcal3.tmp","pcal4.tmp"};
      FILE *fo;

      /* define coordinate range of phase graph */
	  fprintf(plotP->gp,"set title \"%s\"\n",plotP->htitl1);
	  fprintf(plotP->gp,"set xlabel \"%s\"\n",plotP->xtit1);
	  fprintf(plotP->gp,"set ylabel \"%s\"\n",plotP->ytit1);
      fprintf(plotP->gp,"set origin 0.0,0.5\n");
	  fprintf(plotP->gp,"set xrange [%g:%g]\n",plotP->tmin, plotP->tmax);
	  fprintf(plotP->gp,"set yrange [%g:%g]\n",plotP->phsmin, plotP->phsmax);
	  fprintf(plotP->gp,"plot \"%s\" with dots notitle\n",ftemp[0]);
      tdat[0]=(float)(plotP->dtime-plotP->stime);

      if( (fo  = fopen( ftemp[2],"w" )) == NULL ){
         printf(" File (%s) Open Error! \n",ftemp[2]);
         goto exithere;
      } 

      for(n=0; n<plotP->numch; n++){
          w1=plotP->rpcal[n]; 
          w2=plotP->ipcal[n];
          w3=sqrt(w1*w1+w2*w2);
          w4=atan2(-w2,w1)*PIDE;
          ydat1[0]=(float)w4;
          if(n==0){
             fprintf(fo, "%g %g ",tdat[0],ydat1[0]);
          } else {
             fprintf(fo, "%g ",ydat1[0]);
          }
          if(n==plotP->numch-1) fprintf(fo, "\n");
	     // fprintf(plotP->gp,"set label \"%1d\" at first %f, %f center\n",n+1,tdat[0],ydat1[0]);
	     // fprintf(plotP->gp,"unset label\n");  // this is necessary otherwise 
	     // fflush(plotP->gp);

      }
      fclose(fo);
      for(n=0; n<plotP->numch; n++){
         if(n==0){
	         //fprintf(plotP->gp,"plot \"%s\" with points notitle",ftemp[2]);
	         fprintf(plotP->gp,"plot \"%s\" with points title \"CH%1d\"",
                 ftemp[2],n+1);
         } else {
	         //fprintf(plotP->gp,",\"%s\" using 1:%1d with points notitle",
             //    ftemp[2],n+2);
	         fprintf(plotP->gp,",\"%s\" using 1:%1d with points title \"CH%1d\"",
                 ftemp[2],n+2,n+1);
         }
         if(n==plotP->numch-1) fprintf(plotP->gp, "\n");
      }
	  fflush(plotP->gp);


      /* define coordinate range of amp graph */
	  fprintf(plotP->gp,"set title \"%s\"\n",plotP->htitl2);
	  fprintf(plotP->gp,"set xlabel \"%s\"\n",plotP->xtit2);
	  fprintf(plotP->gp,"set ylabel \"%s\"\n",plotP->ytit2);
      fprintf(plotP->gp,"set origin 0.0,0.0\n");
	  fprintf(plotP->gp,"set xrange [%g:%g]\n",plotP->tmin, plotP->tmax);
	  fprintf(plotP->gp,"set yrange [%g:%g]\n",plotP->ampmin, plotP->ampmax);
	  //fprintf(plotP->gp,"plot \"%s\" with dots notitle\n",ftemp[0]);

      if( (fo  = fopen( ftemp[3],"w" )) == NULL ){
         printf(" File (%s) Open Error! \n",ftemp[3]);
         goto exithere;
      } 

      for(n=0; n<plotP->numch; n++){
          w1=plotP->rpcal[n]; 
          w2=plotP->ipcal[n];
          w3=sqrt(w1*w1+w2*w2);
          w4=atan2(-w2,w1)*PIDE;
          ydat2[0]=(float)w3;
          if(n==0){
             fprintf(fo, "%g %g ",tdat[0],ydat2[0]);
          } else {
             fprintf(fo, "%g ",ydat2[0]);
          }
          if(n==plotP->numch-1) fprintf(fo, "\n");
          //fprintf(fo, "%g %g\n",tdat[0],ydat2[0]);
	      //fprintf(plotP->gp,"set label \"%1d\" at first %f, %f center\n",n+1,tdat[0],ydat2[0]);
	      //fprintf(plotP->gp,"unset label\n");  // this is necessary otherwise 
	      //fflush(plotP->gp);
      }
      fclose(fo);
      for(n=0; n<plotP->numch; n++){
         if(n==0){
	         //fprintf(plotP->gp,"plot \"%s\" with points notitle",ftemp[2]);
	         fprintf(plotP->gp,"plot \"%s\" with points title \"CH%1d\"",
                 ftemp[3],n+1);
         } else {
	         //fprintf(plotP->gp,",\"%s\" using 1:%1d with points notitle",
             //    ftemp[2],n+2);
	         fprintf(plotP->gp,",\"%s\" using 1:%1d with points title \"CH%1d\"",
                 ftemp[3],n+2,n+1);
         }
         if(n==plotP->numch-1) fprintf(plotP->gp, "\n");
      }
	  fflush(plotP->gp);
    //  cpgsci(1);  /* black */
exithere:;

#else   // using PGPLOT
      cpgsch((float)1.3);  // set text height 1.3  (default is 1)

      /* define coordinate range of phase graph */
      cpgswin(plotP->tmin, plotP->tmax, plotP->phsmin, plotP->phsmax);
      cpgpanl(1,1);

      tdat[0]=(float)(plotP->dtime-plotP->stime);
      for(n=0; n<plotP->numch; n++){
          w1=plotP->rpcal[n]; 
          w2=plotP->ipcal[n];
          w3=sqrt(w1*w1+w2*w2);
          w4=atan2(-w2,w1)*PIDE;
          ydat1[0]=(float)w4;
          //printf("%f %f\n",tdat[0],ydat1[0]);
          cpgsci(n+1);  // color set
          //cpgpt(1,tdat,ydat1,1);
          cpgpt(1,tdat,ydat1,49+n);
          cpgupdt();
      }
      /* define coordinate range of amp graph */
      cpgswin(plotP->tmin, plotP->tmax, plotP->ampmin, plotP->ampmax);
      cpgpanl(1,2);

      for(n=0; n<plotP->numch; n++){
          w1=plotP->rpcal[n]; 
          w2=plotP->ipcal[n];
          w3=sqrt(w1*w1+w2*w2);
          w4=atan2(-w2,w1)*PIDE;
          ydat2[0]=(float)w3;
          //printf("%f %f\n",tdat[0],ydat2[0]);
          cpgsci(n+1);  // color set
          //cpgpt(1,tdat,ydat1,1);
          cpgpt(1,tdat,ydat2,49+n);
          cpgupdt();
      }
      cpgsci(1);  /* black */
#endif
}

void pcal_plot_after_mode(pcal_plot_t *plotP)
{
      char htitl1[256], xtit1[60], ytit1[60];
      char htitl2[256], xtit2[60], ytit2[60];
      char lab[128],lab1[256],lab2[256],lab3[128],fname[128];
      char cbuf[128];
      float tdat[2], ydat1[2], ydat2[2];
      int k,n,len;
      int hh,mm,ss,seconds;
      double pcalf[4];
      double data[10];
      FILE *fo;
#ifdef USE_GNUPLOT   // using GNUPLOT
      //char *ftemp="pcal.tmp";
      //char ftemp[][20]={"pcal1.tmp","pcal2.tmp","pcal3.tmp","pcal4.tmp"};
      //int npmax;

	   plotP->gp= popen(GNUPLOT_EXE,"w");
       if(plotP->gp==NULL) goto exithere;
       if(strncmp(plotP->g_device,G_PS_DEVICE,strlen(G_PS_DEVICE))==0){ // postscript set
	      //fprintf(gp,"set terminal post eps color\n");
	      fprintf(plotP->gp,"set terminal post landscape 12 color\n");
	      fprintf(plotP->gp,"set output \"%s\"\n",plotP->ps_file);
       } else {
	      fprintf(plotP->gp,"set terminal %s\n",plotP->g_device);
       }
	   fflush(plotP->gp);
	   fprintf(plotP->gp,"set multiplot\n");
	   fprintf(plotP->gp,"set size 1.0,0.5\n");

#else   // using PGPLOT
      if(cpgopen(plotP->g_device) < 1) goto exithere;
      if(strncmp(plotP->g_device,"/GW",3)==0) cpgeras();   // erase just in case 2007-10-26

      if(strncmp(plotP->g_device,"/XWIN",5)==0 ||
         strncmp(plotP->g_device,"/XSER",5)==0 ) { /* set background white 
                                                   and text black */
         cpgscr(0,1.0,1.0,1.0);  /* Background color */
         cpgscr(1,0.0,0.0,0.0);  /* Foreground color */
      }
      cpgsubp(1,2);  // divide plot pannel into 1X2
#endif

      pcal_info_get(plotP);

      // set title

      strcpy(fname,plotP->fname);
      len=strlen(fname);
      for(n=0; n<len; n++){  // convert '\' to '/' to avoid misoperation in plot
          if(fname[n]=='\\') fname[n]='/';
      }

      // set PCAL frequency in kHz
      for(n=0; n<plotP->numch; n++) pcalf[n]=plotP->pcalf[n]/1000.0;
      if(plotP->numch==1){
         sprintf(lab,"Ch#1: %.1fkHz\0",pcalf[0]);
      } else {
         sprintf(lab,"Ch#1: %.1fkHz  Ch#2: %.1fkHz  Ch#3: %.1fkHz  Ch#4: %.1fkHz\0",
          pcalf[0],pcalf[1],pcalf[2],pcalf[3]);
      }

      // set start time

      seconds=(int)plotP->stime;
      hh=seconds/3600;
      mm=(seconds % 3600)/60;
      ss=seconds % 60;

      if(plotP->hd.format==2){ // VSSP32 
         sprintf(lab1,"Start Time(UT) %04d/%03d %02d:%02d:%02d   \0",
             plotP->hd.year,plotP->hd.totalday,hh,mm,ss);
      } else {
         sprintf(lab1,"Start Time(UT) %02d:%02d:%02d   \0",
            hh,mm,ss);
      }
      plotP->hh=hh; plotP->mm=mm; plotP->ss=ss;

      strcat(lab1,lab);


      //set sampling parameters
      if(plotP->hd.format==2){ // VSSP32 
          sprintf(lab2,"VSSP32  Fs=%.1fMHz AD=%2d #_of_CH=%1d  LPF=%dMHz  \0",
              plotP->hd.sfreq/1000.0,plotP->hd.adbit,
              plotP->hd.numch,plotP->hd.filter);
      } else {
          sprintf(lab2,"VSSP  Fs=%.1fMHz AD=%2d #_of_CH=%1d  \0",
              plotP->hd.sfreq/1000.0,plotP->hd.adbit,
              plotP->hd.numch);
      }

      // set station name, host PC info

      lab3[0]='\0';
      if(plotP->hd.format==2){ // VSSP32 
         sprintf(lab3," Station: %s  Host PC: %s\0",
            plotP->hd.aux_station, plotP->hd.aux_hostname);
      }


      strcat(lab2,lab3);

      // set title axis label for PLOT#1
      sprintf(htitl1,"PCAL PHASE (%s)\0",fname);
      strcpy(xtit1,"TIME (sec)");
      strcpy(ytit1,"PHASE (deg)");

      // set title axis label for PLOT#2
      sprintf(htitl2,"PCAL AMPLITUDE (%s)\0",fname);
      strcpy(xtit2,"TIME (sec)");
      strcpy(ytit2,"AMPLITUDE");


      //for(n=0; n<4; n++){
      /* define coordinate range of phase graph */
      plotP->tmin=0.0;
      plotP->tmax=(float)plotP->scanlength;
      plotP->phsmin=-180.0;
      plotP->phsmax=180.0;
      /* define coordinate range of amp graph */
      plotP->ampmin=0.0;
      plotP->ampmax=(float)0.15;

#ifdef USE_GNUPLOT   // using GNUPLOT


      // PLOT#1   phase
      fprintf(plotP->gp,"set origin 0.0,0.5\n");
	  fprintf(plotP->gp,"set xrange [%g:%g]\n",plotP->tmin, plotP->tmax);
	  fprintf(plotP->gp,"set yrange [%g:%g]\n",plotP->phsmin, plotP->phsmax);
      fprintf(plotP->gp,"set ytics -180, 90\n");
      
      /* label the axes */
	  fprintf(plotP->gp,"set xlabel \"%s\"\n",xtit1);
	  fprintf(plotP->gp,"set ylabel \"%s\"\n",ytit1);

      if(strncmp(plotP->g_device,"windows",7)==0 || 
          strncmp(plotP->g_device,"x11",3)==0 ){ // display
	     fprintf(plotP->gp,"set title \"%s\"\n",htitl1);
	     fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.14 center\n",lab1);
	     fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.05 center\n",lab2);
      } else { // postscript
	     fprintf(plotP->gp,"set title \" \"\n");  // this is to keep dummy space
	     fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.14 center\n",htitl1);
	     fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.08 center font \"Helvetica,10\"\n",lab1);
	     fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.03 center font \"Helvetica,10\"\n",lab2);
      }
      for(n=0; n<plotP->numch; n++){
         if(n==0){
	         fprintf(plotP->gp,"plot \"%s\" using 1:3 with points title \"CH%1d\"",
                 plotP->pfile,n+1);
         } else {
	         //fprintf(plotP->gp,",\"%s\" using 1:%1d with points notitle",
             //    ftemp[2],n+2);
	         fprintf(plotP->gp,",\"%s\" using 1:%1d with points title \"CH%1d\"",
                 plotP->pfile,2*n+3,n+1);
         }
         if(n==plotP->numch-1) fprintf(plotP->gp, "\n");
      }
	  fflush(plotP->gp);



      // PLOT#2  amp
	  fprintf(plotP->gp,"unset label\n");  // this is necessary otherwise 

      fprintf(plotP->gp,"set origin 0.0,0.0\n");
	  fprintf(plotP->gp,"set xrange [%g:%g]\n",plotP->tmin, plotP->tmax);
	  fprintf(plotP->gp,"set yrange [%g:%g]\n",plotP->ampmin, plotP->ampmax);
      fprintf(plotP->gp,"set ytics autofreq\n");
      
      /* label the axes */
	  fprintf(plotP->gp,"set xlabel \"%s\"\n",xtit2);
	  fprintf(plotP->gp,"set ylabel \"%s\"\n",ytit2);
      if(strncmp(plotP->g_device,"windows",7)==0 || 
          strncmp(plotP->g_device,"x11",3)==0 ){ // display
	     fprintf(plotP->gp,"set title \"%s\"\n",htitl2);
	     fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.14 center\n",lab1);
	     fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.05 center\n",lab2);
      } else { // postscript
	     //fprintf(plotP->gp,"set title \"%s\"\n",htitl2);
	     fprintf(plotP->gp,"set title \" \"\n");  // this is to keep dummy space
	     fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.14 center\n",htitl2);
	     fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.08 center font \"Helvetica,10\"\n",lab1);
	     fprintf(plotP->gp,"set label \"%s\" at graph 0.5, 1.03 center font \"Helvetica,10\"\n",lab2);
      }

      for(n=0; n<plotP->numch; n++){
         if(n==0){
	         fprintf(plotP->gp,"plot \"%s\" using 1:2 with points title \"CH%1d\"",
                 plotP->pfile,n+1);
         } else {
	         //fprintf(plotP->gp,",\"%s\" using 1:%1d with points notitle",
             //    ftemp[2],n+2);
	         fprintf(plotP->gp,",\"%s\" using 1:%1d with points title \"CH%1d\"",
                 plotP->pfile,2*n+2,n+1);
         }
         if(n==plotP->numch-1) fprintf(plotP->gp, "\n");
      }
	  fflush(plotP->gp);
 	  fprintf(plotP->gp,"unset label\n");  // this is necessary otherwise 
	  fflush(plotP->gp);
	  //getchar();  // this is necessary to keep display otherwise deleted

#else   // using PGPLOT

      /* phase plot */
      cpgsch(2.0);  // set text height 2.0  (default is 1)
      cpgslw(2);    // set line width
      cpgsci(1);  /* black */
      //cpgenv(plotP->tmin, plotP->tmax, plotP->phsmin, plotP->phsmax, 0, 0);   // linear scale
      cpgenv(plotP->tmin, plotP->tmax, plotP->phsmin, plotP->phsmax, 0, -2);   // linear scale
      // cpgenv will move window automatically to next
      
      /* label the axes */
      cpgpanl(1,1);
      cpgbox("BCNST",0.0,0,"BCNST",90.0,3);  // to improve Y axis appearance
      //cbuf[0]='N';
      //cpgaxis(cbuf,plotP->tmin, plotP->phsmin,plotP->tmin, plotP->phsmax,
      //   plotP->phsmin,plotP->phsmax,90.0,3,0.0,1.0,0.5,-0.5,0.0);
      cpglab(xtit1,ytit1,htitl1);
      /* sub label write*/
      cpgsch(1.5);
      cpgmtxt("T",1.5,0.5,0.5,lab1);
      cpgmtxt("T",0.5,0.5,0.5,lab2);

      cpgsch((float)1.3);

      if( (fo = fopen(plotP->pfile, "r" )) == NULL ){
            printf("pcal_plot_after_mode: File (%s) Open Error! \n",
                plotP->pfile);
		    goto exithere;
      }

      while(fgets(cbuf, 128, fo) != NULL){
         if(cbuf[0]=='#') continue;  // this is comment line (nothing to do) 
         k=parseline2item(cbuf,data);
         tdat[0]=(float)data[0];
         for(n=0; n<plotP->numch; n++){
            ydat1[0]=(float)data[2*n+2];
            cpgsci(n+1);  // color set
            cpgpt(1,tdat,ydat1,49+n);
            cpgupdt();
         }
      }

      /* amplitude plot */
      cpgsch(2.0);  // set text height 2.0  (default is 1)
      cpgslw(2);    // set line width
      cpgsci(1);  /* black */
      /* define coordinate range of amp graph */
      /* label the axes */
      cpgenv(plotP->tmin, plotP->tmax, plotP->ampmin, plotP->ampmax, 0, 0);   // linear scale
      cpgpanl(1,2);
      //sprintf(htitl,"PCAL AMPLITUDE (%s)\0",fname);
      //strcpy(ytit,"AMPLITUDE");
      cpglab(xtit2,ytit2,htitl2);
      /* sub label write*/
      cpgsch(1.5);
      cpgmtxt("T",1.5,0.5,0.5,lab1);
      cpgmtxt("T",0.5,0.5,0.5,lab2);
      cpgsch((float)1.3);

      rewind(fo);
      while(fgets(cbuf, 128, fo) != NULL){
         if(cbuf[0]=='#') continue;  // this is comment line (nothing to do) 
         k=parseline2item(cbuf,data);
         tdat[0]=(float)data[0];
         for(n=0; n<plotP->numch; n++){
            ydat2[0]=(float)data[2*n+1];
            cpgsci(n+1);  // color set
            cpgpt(1,tdat,ydat2,49+n);
            cpgupdt();
         }
      }
      fclose(fo);
      cpgsci(1);  /* black */

#endif
      pcal_plot_end(plotP);
      if(strncmp(plotP->g_device,G_PS_DEVICE,strlen(G_PS_DEVICE))==0){ // postscript
          printf("Created PostScript File ==> %s\n",plotP->ps_file);
      }
exithere:;

}


//void pcal_engine(xros_engine_param_t *rp, char *fname)   // Version 2007-10-22
void pcal_engine(run_param_pcalcheck_t *runp)   // Version 2007-10-22
// PCAL check
{
    int sfreq, adbit, numch, kposx;
    int nspp;
    int n_usampl;
    int i,imax, m,kaisu,n;
    int kpp, kppold;
    int usampl;
    int smode;
    double tinteg;
    long int samplpp;  // samples/PP
    unsigned char *ibufx;
    int bits, bytes, numb;
    int scount[4];
    float fsampl,tsampl;
    int *idatax;
    unsigned char *bdatax;
    double dtimex;
    //char lab2[60], iplab[40], ltime1st[60];
    ip_vlbi_file_t X, Xs;   // IP-VLBI data type variables  Xs for save start info
	double telapsed;          // processing time
	time_t tstart, tfinish;   // to measure processing time
    int timecheck_mode;       // Data time check mode
                              //   0: no data time check (used for auto correlation mode)
                              //   1: data time check    (for usual cross correlation mode)
    int delayed_valid_flag;   // 1:no sync error during current PP processing
                              // 0:sync error occurred during current PP processing
    double w0,w1,w2,w3,w4;
    int cpat[256];
    char pcalcT[65536];
    int64_t fsize;

    double rpcalsum[4],ipcalsum[4];
    int mode_1bit;  // 1: 1 bit sampoling mode
    int headerbits;  // header bits
    double scanlength;
    double t1pp;
    double dtime_bopp;
    double dtime_mopp;
    double pide=PIDE;
    const int cbyte=256;      // for temporary PP data file read
    char *cbuf;               // for temporary PP data file read
    FILE *fo;            // for pcal results out
    //	ppdata_tv3 pp;          // pp data  type ver 3.0
    pcaltype_tv3 pcx;    // for PCAL detection type ver 3.0
    pcal_plot_t plotP;

    printf("pcal_engine:  %s\n",PROGRAM_VERSION);
    //printf("pcal_engine:  Ver. 2007-11-04\n");


//    hostnameget(hostname);



    strcpy(plotP.g_device,runp->pgdevice);
    strcpy(plotP.fname,runp->fname);
    strcpy(plotP.ps_file,runp->ps_file);

    /* check given file is data file or PCAL results file */
    if(pcal_info_file_check(runp->fname)==0){  // this is PCAL results file
        strcpy(plotP.pfile,runp->fname);
        runp->after=1;
        goto after_plot;
    }

	cbuf=(char *)malloc(cbyte); 
    strcpy(plotP.pfile,runp->pfile);

    pcx.chsize=4;

	setctabl0(cpat);   // bit 1 (in a byte) counter setup
    setctabl(pcalcT);  // setup correlation table for PCAL detection 
    for(n=0; n<4; n++){
        rpcalsum[n]=0.0;
        ipcalsum[n]=0.0;
    }

    mode_1bit=0;   // 1: data are 1bit sampling mode

    /*  memory for PP and PCAL data allocation */
    pcx.pcalf=(double *)malloc(pcx.chsize*sizeof(double));
    pcx.rpcal=(double *)malloc(pcx.chsize*sizeof(double));
    pcx.ipcal=(double *)malloc(pcx.chsize*sizeof(double));
    pcx.sum1=(double *)malloc(pcx.chsize*sizeof(double));
    pcx.sum2=(double *)malloc(pcx.chsize*sizeof(double));

    /* initialize PCAL data variables */
    inipcaldatv3(&pcx);

    for(n=0; n<4; n++) pcx.pcalf[n]=runp->pcalf[n];
    pcx.mode=1;
    for(n=0; n<3; n++) {
       for(m=n+1; m<4; m++) {
           if(pcx.pcalf[n]!=pcx.pcalf[m]) pcx.mode=2;
       }
    }

//    printf("Debug pcx.mode = %d\n",pcx.mode);
//    printf("Debug pcalf[0] = %f\n",pcx.pcalf[0]);
//    printf("Debug pcalf[1] = %f\n",pcx.pcalf[1]);
//    printf("Debug pcalf[2] = %f\n",pcx.pcalf[2]);
//    printf("Debug pcalf[3] = %f\n",pcx.pcalf[3]);

    /*********************************************/
    /* set time_check mode                       */
    /*    default is 1 (cross correlation mode)  */
    /*    if year of prt was set to 0,           */
    /*    time_check mode set to 0 for           */
    /*    auto correlation processing            */
    /*********************************************/

    timecheck_mode=0;

    /* initialize PP data variables */
//    inippdatv3(&pp);

    /* set PCAL detection mode */

    //pcx.mode=1;  // pcal detection on with all same frequency mode
    //pcx.tmode=0;  // pcal detection every 1 sec mode
    pcx.tmode=1;  // pcal detection every PP mode


    t1pp=1.0;
//    t1pp=rp->runp.t1pp;
//    sekibun=rp->runp.sekibun;
//    pp.t1pp=t1pp;         // information save for result output
//    pp.sekibun=sekibun;   // information save for result output

    /******************************************/
    /* set obs start time in seconds in day   */
    /******************************************/

//    startsec=(long int)(((long int)rp->ape.istart[2])*3600.0
//		     +(long int)rp->ape.istart[3]*60.0
//			 +(long int)rp->ape.istart[4]);
//    startsec+=rp->runp.startoffset;



    //X.fname=rp->ape.xfile;      // X data file name
    X.fname=runp->fname;      // X data file name

    X.vflag=1;                  // X valid data flag set
    X.vflag_old=1;              // X valid old data flag set
    delayed_valid_flag=1;       // Current PP validity flag


//loopentry: 


        /*********************************************/
        /*  obs data file open                       */
        /*********************************************/

        fsize=get_file_size(X.fname);  // get file size in bytes
	    if(fsize == -1){
            printf("File %s not found! \n",X.fname);
            goto exitend2;
        }
        //printf("Debug file size %lld\n",fsize);

        printf("pcal_engine: Data File is %s\n",X.fname);
        if( (X.stream  = fopen( X.fname, "rb" )) == NULL ){
            printf("xros_engine: File Open Error! \n");
		    goto exitend2;
        }

        /****************************************************************/
        /* Data header check to get sampling parameters  and start time */
        /****************************************************************/

        X.hd.format=0;  // for 1st call  2006-01-03
        if(checkheader(&X) < 0){
            goto exitend;
        }
        Xs=X;

        /********************************/
        /* set internal-use parameters  */
        /********************************/

        headerbits=64;
        if(X.hd.format==2) headerbits=256; // for VSSP32

        sfreq=Xs.sfreq;       // sampling frequency (kHz)
        adbit=Xs.adbit;       // A/D bits
        numch=Xs.numch;       // # of channel in data (1 or 4)
        fsampl=(float)(1000.0*(float)sfreq);   // sampling frequency (Hz)
        tsampl=(float)(1.0/fsampl);            // sampling period (sec)
        bits=((long)fsampl)*adbit*numch;  // # of bits in 1 sec data except header
        bytes=bits/8;                     // # of bytes in 1 sec data except header

        scanlength=(double)fsize/((double)(bytes+headerbits/8)); //scan length in sec

        pcx.numch=Xs.numch;               // to keep # of ch info in PCAL type data

        X.fsample=fsampl;                 // to keep fsampl info in PPDATA type data

        X.tsample=tsampl;                 // to keep tsampl info in PPDATA type data

//        pp.adbit=adbit;                   // information save for output
//        pp.numch=numch;                   // information save for output
//        pp.fsample=fsampl;                // information save output
       
        samplpp=(long int)(fsampl*t1pp);  // # of samples in a PP

        if(adbit==1) mode_1bit=1;   // data are 1bit sampling mode

        /*****************************************/
        /*  set unit # of samples to process     */
        /*****************************************/

            //   smode  : search mode   =0  wide search range mode
            //                          =1  middle range
            //                          =2  narrow range

        smode=0;
        if(smode ==0){ usampl=200000; }
        else if (smode ==1) {
           usampl=10000;
        } else {
           usampl=1000;      // for smode=2
        }

        //    check usampl condition           added on 2003-02-09
        //    because usampl*adbit*numch/32 must be an integer

        if((usampl*adbit*numch)%32 != 0) {
           while ((usampl*adbit*numch)%32 != 0) {
             usampl=2*usampl;
           }
           printf("pcal_engine:  new usampl %d\n",usampl);
        }
        if(samplpp%usampl != 0){
            printf("pcal_engine:  samplpp,usampl condition error!  %d %d\n",samplpp,usampl);
            printf("pcal_engine:  Please change smode and run again!\n");
            goto exitend;
        }


        nspp=samplpp/usampl;             // # of usampls in a PP
        imax=(sfreq*1000)/usampl;        // # of usampls in a second
        numb=usampl*adbit*numch/8;       // # of bytes in a usampl

        X.unitp=(double)usampl*tsampl;   // period in a usampl (sec)


        if(X.hd.format == 1) printf("pcal_engine:  << VSSP format\n");
        if(X.hd.format == 2) printf("pcal_engine:  << VSSP32 format  (%2d.%02d)\n",
                           X.hd.v_major,X.hd.v_minor);
//        printf("pcal_engine:  << runmode in xros_engine (runmode)= %d\n",runmode);
//        printf("pcal_engine:  << smode in xros_engine (smode)    = %d\n",smode);
        printf("pcal_engine:  << # of samples in a unit (usampl) = %d\n",usampl);
//        printf("pcal_engine:  << FFT size for processing         = %d\n",nfft);  // debug
        printf("pcal_engine:  << PP period in sec (t1pp)         = %.1f\n",t1pp);  // debug
        printf("pcal_engine:  << # of usampl  in a PP (nspp)     = %d\n",nspp);
        printf("pcal_engine:  << # of usampl  in 1 sec (imax)    = %d\n",imax);
        printf("pcal_engine:  << # of bytes   in a usampl (numb) = %d\n",numb);
        printf("pcal_engine:  << File size in bytes              = %lld\n",fsize);
        printf("pcal_engine:  << Supposed scan length (sec)      = %f\n",scanlength);

        if(Xs.hd.format==2){ // VSSP32 
           printf("pcal_engine: Start Time(UT) %04d/%03d %02d:%02d:%02d\n",
              Xs.hd.year,Xs.hd.totalday,Xs.hh,Xs.mm,Xs.ss);
           printf("pcal_engine: Sampler Host PC :  %s\n",Xs.hd.aux_hostname);
           if(Xs.hd.aux_fmt==1)  
               printf("pcal_engine: Sampler Station :  %s\n",Xs.hd.aux_station);
        }
        /*********************************************/
        /*  internal use array memory allocation     */
        /*********************************************/

        ibufx  = (unsigned char *)malloc(numb);                // X data read buffer
        idatax = (int *)malloc(numch*usampl*sizeof(int));      // X decoded each channel data 
        bdatax = (unsigned char *)malloc(numch*usampl/8);      // X decoded each channel data 

        /**************************/
        /* initialize counter etc */
        /**************************/

//        sprintf(ltime1st,"Time(UT) %02d:%02d:%02d\0",Xs.hh,Xs.mm,Xs.ss);

		X.imax=imax;
		X.icnt=-1;
        for(m=0; m<numch; m++) scount[m]=0;     // reset # of usampl data in a PP by channel

        //for(k=0; k<numch*nfft; k++){spows[k]=0.0;}  //not used

        kaisu=0;      // total # of usampl data in an OBS
        kpp=0;        // PP # counter
		kppold=-1;    // PP # old number
        kaisu=0;      // total # of usampl data in an OBS
        n_usampl=0;   // # of usampl data in a PP        
        tinteg=0.0;   // actual integration time

        /************************************/
		/* X data 1st one usampl data read  */
        /************************************/
		//if (Dat1unitRead_w_Pcdetv2(&X,ibufx,idatax,numb,&kposx,&pcx) < 0) goto exitend;
        if(mode_1bit == 1){
           if(DatRead1bitSv(&X, ibufx, bdatax, numb, &pcx, cpat, pcalcT)<0) goto exitend;
        } else {
     	   if (Dat1unitRead_w_Pcdetv3(&X,ibufx,idatax,numb,&kposx,&pcx) < 0) goto exitend;
        }
	    //printf("n_usampl,X.icnt, X.dtime %d %d %f\n",n_usampl,X.icnt,X.dtime);  //debug
        dtimex=X.dtime;
		printf("pcal_engine:   dtimex %f\n",dtimex); // debug

        printf("pcal_engine: Start X data time : %f\n",dtimex);
        Xs.dtime=X.dtime;

        tstart= time(NULL);  // to measure processing time for each PP 


        /************************************/
		/* PCAL results output file open    */
        /************************************/

        if( (fo  = fopen(runp->pfile, "w" )) == NULL ){
            printf("pcal_engine:   results out file (%s) Open Error! \n",runp->pfile);
            goto exitend;
        }
        printf("pcal_engine:   pcal results out file (%s) Opened! \n",runp->pfile);

	    fprintf(fo,"# PCAL detected by pcalcheck %s\n",PROGRAM_VERSION);
	    fprintf(fo,"# File : %s\n",runp->fname);
        if(Xs.hd.format==2){ // VSSP32 
           fprintf(fo,"# Sampler Type     :  K5/VSSP32 %2d.%02d\n",
               Xs.hd.v_major,Xs.hd.v_minor);
           fprintf(fo,"# Start Time(UT)   :  %04d/%03d %02d:%02d:%02d\n",
              Xs.hd.year,Xs.hd.totalday,Xs.hh,Xs.mm,Xs.ss);
           fprintf(fo,"# Scan Length(sec) :  %.2f\n",scanlength);
           fprintf(fo,"# Sampler Host PC  :  %s\n",Xs.hd.aux_hostname);
           if(Xs.hd.aux_fmt==1)  
               fprintf(fo,"# Sampler Station  :  %s\n",Xs.hd.aux_station);
           fprintf(fo,"# Fs= %.1f MHz AD= %2d #_of_CH= %1d  LPF= %d MHz\n",
              Xs.hd.sfreq/1000.0,Xs.hd.adbit,
              Xs.hd.numch,Xs.hd.filter);
        } else {
           fprintf(fo,"# Sampler Type     :  K5/VSSP\n");
           fprintf(fo,"# Start Time(UT)   :  %02d:%02d:%02d\n",
             Xs.hh,Xs.mm,Xs.ss);
           fprintf(fo,"# Scan Length(sec) :  %.2f\n",scanlength);
           fprintf(fo,"# Fs= %.1f MHz AD= %1d #_of_CH= %1d\n",
              Xs.hd.sfreq/1000.0,Xs.hd.adbit,
              Xs.hd.numch);
        }

        if(pcx.numch==1){
           fprintf(fo,"# CH #1: %.1f kHz\n",pcx.pcalf[0]/1000.0);
        } else {
           fprintf(fo,"# CH #1: %.1f kHz  CH #2: %.1f kHz  CH #3: %.1f kHz  CH #4: %.1f kHz\n",
            pcx.pcalf[0]/1000.0,pcx.pcalf[1]/1000.0,pcx.pcalf[2]/1000.0,pcx.pcalf[3]/1000.0);
        }
        fprintf(fo,"#     ");
        for(m=0; m<pcx.numch; m++) fprintf(fo,"               CH #%1d",m+1);
        fprintf(fo,"\n");
        fprintf(fo,"# time(sec)");
        for(m=0; m<pcx.numch; m++) fprintf(fo,"     Amp  Phase(deg)");
        fprintf(fo,"\n");

        plotP.scanlength=scanlength;
        plotP.stime=Xs.dtime;
        plotP.numch=pcx.numch;
        for(m=0; m<pcx.numch; m++){
           plotP.pcalf[m]=pcx.pcalf[m];
           plotP.rpcal[m]=pcx.rpcal[m];
           plotP.ipcal[m]=pcx.ipcal[m];
        }
        memcpy(&plotP.hd,&Xs.hd,sizeof(Xs.hd));
        plotP.tinteg=tinteg;

        //dtime_bopp=Xs.dtime;

        if(runp->after==0){
           // plot start
           pcal_plot_start(&plotP);

        }

   /****************************************************/
   /*        unit read and PP loop start point         */
   /****************************************************/
loop1:;   

 
         kaisu++;               // total # of usampl data count up
         n_usampl++;            // # of usampl data in a PP count up
         tinteg += X.unitp;     // integration period update

         if(n_usampl==1) dtime_bopp=X.dtime;
         if(n_usampl== nspp/2+1) dtime_mopp=X.dtime;

         /*********************************/
         /*  PP data save (write to file) */
         /*********************************/

         if(n_usampl == nspp){ //  pp data save
	         //printf("X.icnt, X.dtime %d %f\n",X.icnt,X.dtime);  //debug

             /* print processing time for each PP */

             tfinish= time(NULL);                   
             telapsed = difftime(tfinish,tstart);

             /* normalizing PCAL detection data */

             if(mode_1bit == 1){
                    pcaldatnorm_1bit(&pcx);
                    //for(m=0; m< pp.numch; m++) printf("pcx %f %f\n",pcx.rpcal[m],pcx.ipcal[m]);
                    /* PCAL write */
	                //printf("X-PCAL\n");
                    //for(m=0; m<pcx.numch; m++){
                    //   w1=pcx.rpcal[m]; 
                    //   w2=pcx.ipcal[m];
                    //   w3=sqrt(w1*w1+w2*w2);
                    //   w4=atan2(-w2,w1)*PIDE;
                    //   printf("%d %d %f %f %f %f\n",m,pcx.ns,w1,w2,w3,w4);  
                    //}
         //       }
             } else {
                pcaldatnormv3(&pcx);

                /* PCAL write */
	            //printf("X-PCAL\n");
                for(m=0; m<pcx.numch; m++){
                   w0=sqrt(pcx.sum2[m]-pcx.sum1[m]*pcx.sum1[m]/(double)pcx.ns);
                   w1=pcx.rpcal[m]/w0; 
                   w2=pcx.ipcal[m]/w0;
                   w3=sqrt(w1*w1+w2*w2);
                   w4=atan2(-w2,w1)*PIDE;
                   //printf("%d %d %f %f %f %f\n",m,pcx.ns,w1,w2,w3,w4);
                   pcx.rpcal[m]=w1;
                   pcx.ipcal[m]=w2;
                }
             }

             /* detected PCAL  */
             //fprintf(fo," %f  ",dtime_bopp-Xs.dtime);
             fprintf(fo," %f  ",dtime_mopp-Xs.dtime);


             printf("pcal_engine:  << Detected PCAL >>\n");
             printf("pcal_engine:   CH  samples  Freq(kHz)    Amp  Phase(deg)\n");
             for(m=0; m<pcx.numch; m++){
                 w1=pcx.rpcal[m]; 
                 w2=pcx.ipcal[m];
                 w3=sqrt(w1*w1+w2*w2);
                 w4=atan2(-w2,w1)*PIDE;
                 printf("pcal_engine:   %2d %8d %8.1f  %8.6f %8.2f\n",
                     m,pcx.ns,pcx.pcalf[m]/1000.0,w3,w4);

                 fprintf(fo,"   %8.6f %8.2f",w3,w4);
                 plotP.rpcal[m]=pcx.rpcal[m]; 
                 plotP.ipcal[m]=pcx.ipcal[m]; 
             }
             fprintf(fo,"\n");


             // plot PP data
             //plotP.dtime=X.dtime;
             //plotP.dtime=dtime_bopp;
             plotP.dtime=dtime_mopp;

             if(runp->after==0){
                pcal_plot_point(&plotP);
             }


             /* PP data correlation data etc save (write) to file */ 

             //out_for_pvwave_pp(fo, csum, scount, nfft, &pp, &pcx, &pcy, delayed_valid_flag);
             //out_for_pvwave_pp3(fo, csum, scount, nfft, &pp, &pcx, &pcy, delayed_valid_flag);
             //printf("pcal_engine: PP# %d data saved\n",kpp+1);
             printf("pcal_engine:   Time elapsed for 1PP processing is %f sec\n",telapsed);

             /*********************/
             /* reset counter etc */
             /*********************/

             /* clear pcal count etc*/
             if(pcx.tmode == 1){   // PCAL detection every PP mode  otherwise every 1 sec mode
                 for(m=0; m<pcx.numch; m++){
                    rpcalsum[m]+=pcx.rpcal[m]; 
                    ipcalsum[m]+=pcx.ipcal[m]; 

                    pcx.rpcal[m]=0.0;
                    pcx.ipcal[m]=0.0;
                    pcx.sum1[m]=0.0;
                    pcx.sum2[m]=0.0;
                 }
                 pcx.ns=0;
                 pcx.kpp++;
             }


             /* counter flag reset */

             delayed_valid_flag=1;// PP data validity flag set
             n_usampl=0;          // # of usampl data counter
             kpp++;               // PP# counter update
             tstart= time(NULL);  // PP processing initial CPU time get

             /* integration period check */

             //printf("pcal_engine:   X data time (BOPP) : %f\n",X.dtime); //debug
             printf("pcal_engine:   X data time (BOPP) : %f\n",dtime_bopp); //debug
//             if(sekibun > 0.0) {
//                 printf("pcal_engine:   processed data (%.1f/%.1f)\n",tinteg,sekibun); //debug
//             } else {
//                 printf("pcal_engine:   processed data (%.1f)\n",tinteg); //debug
//             }
//
             if( tinteg+0.5*X.unitp >= runp->sekibun && runp->sekibun > 0.0) goto exitend;  // quit correlation
         }

    
		 //if (Dat1unitRead_w_Pcdetv2(&X,ibufx,idatax,numb,&kposx,&pcx) < 0) goto exitend;
         if(mode_1bit == 1){
            if(DatRead1bitSv(&X, ibufx, bdatax, numb, &pcx, cpat, pcalcT)<0) goto exitend;
         } else {
  	        if (Dat1unitRead_w_Pcdetv3(&X,ibufx,idatax,numb,&kposx,&pcx) < 0) goto exitend;
         }
         dtimex=X.dtime;

		 goto loop1;

      /*************************************/
      /*   Bottom of unit read and PP loop */
      /*************************************/


//    exitloop:;
 

 exitend: ;

       /* data file close */

       i=fclose( X.stream ) ;    // X data file

       i=fclose( fo ) ;          // results out file

       /* PCAL summary plot */
       printf("\n\n");
       printf("==================  PCALCHECK SUMMARY  ==================\n");
       printf(" FILE NAME = %s\n",plotP.fname);
       //printf(" EXP.CODE = %s    OBS# = %d\n",hd->excode,hd->nobs);
       //printf(" BASELINE  = %s - %s   (%s)\n",hd->statx,hd->staty,hd->lbase);
       //printf(" SOURCE    = %s\n",hd->srcnam);
       printf(" START TIME= %4dY%03dD%02dh%02dm%02dsec\n",
          plotP.hd.year,plotP.hd.totalday,plotP.hh,plotP.mm,plotP.ss);
       printf(" SCAN LENGTH  (sec)  = %6.1f\n",plotP.scanlength);
       printf(" ACCUM. PERIOD(sec)  at pcalcheck  = %6.1f\n",plotP.tinteg);
       if(plotP.hd.format==2){ // VSSP32 
          printf(" SAMPLER : K5/VSSP32  Fs=%.1fMHz  AD=%2d  #CH=%1d  LPF=%dMHz\n",
              plotP.hd.sfreq/1000.0,plotP.hd.adbit,
              plotP.hd.numch,plotP.hd.filter);
       } else {
          printf(" SAMPLER : K5/VSSP  Fs=%.1fMHz AD=%2d #_of_CH=%1d\n",
              plotP.hd.sfreq/1000.0,plotP.hd.adbit,
              plotP.hd.numch);
       }
       printf("\n");
       printf("  CH#  PCAL FREQ.(kHz)      PHASE(deg.)     AMPLITUDE\n");
       printf(" -------------------------------------------------------\n");
       for(m=0; m<pcx.numch; m++){
          w1=rpcalsum[m]/(double)pcx.kpp; 
          w2=ipcalsum[m]/(double)pcx.kpp;
          w3=sqrt(w1*w1+w2*w2);
          w4=atan2(-w2,w1)*PIDE;
          printf("%4d %10.1f  %20.2f %14.4f\n",
                     m+1,pcx.pcalf[m]/1000.0,w4,w3);

       }
       printf(" -------------------------------------------------------\n");
       printf("=========================================================\n");
       printf("\n\n");
       printf("Pcalcheck results are stored to \"%s\"\n",plotP.pfile);
       printf("You can execute \"pcalcheck %s\" to get fast re-plot.\n",plotP.pfile);
       printf("\n\n");

       /* memory release */
       free(ibufx);
       free(idatax);
       free(cbuf);
       free(pcx.pcalf);
       free(pcx.rpcal);
       free(pcx.ipcal);
       free(pcx.sum1);
       free(pcx.sum2);

       if(runp->after==0) pcal_plot_end(&plotP);

after_plot:

       /* after plot (if selected so) and close the graphics device */
       if(runp->after==1) pcal_plot_after_mode(&plotP);


//       rp->loopcounter++;

//       if(rp->runp.loop_param == 1) goto loopentry;

 exitend2: ;
}

int main(int argc, char *argv[])
{
	char* ofile=PCALCHECK_OUT_DFLT;    // pcal detection results out put file
                                   // GNUPLOT data format
    run_param_pcalcheck_t runp;
    double telapsed;           // to measure processing time
	time_t tstart, tfinish;    //           "
    pcal_plot_t plotP;

    /* run parameters info */
	if(argc == 1){
       print_usage();
       exit(0);
    }
    strcpy(runp.pfile,ofile);
    run_param_get_4_pcalcheck(argc, argv, &runp);

    /* check given file is data file or PCAL results file */
    if(pcal_info_file_check(runp.fname)==0){  // this is PCAL results file
        strcpy(plotP.ps_file,runp.ps_file);
        strcpy(plotP.g_device,runp.pgdevice);
        strcpy(plotP.pfile,runp.fname);
        pcal_plot_after_mode(&plotP);

    } else {  // this raw data file

        tstart= time(NULL);

        /*  do pcal detection  */

        pcal_engine(&runp);          

        tfinish= time(NULL);
        telapsed = difftime(tfinish,tstart);
        printf("Time elapsed (sec) for One obs process is %f\n",telapsed);
    }

    return 0;
}
